/*   IRC Services2
*   Copyright (c) 2003 Services Developers
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of version 1 of the GNU General Public License,
*  as published by the Free Software Foundation.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. The GNU General Public License
*  contains more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* $Id: aclone.c,v 1.16 2003/09/27 16:14:39 wcampbel Exp $
*/

#include "irc_string.h"
#include "defines.h"
#include "clone.h"
#include "privs.h"
#include "send.h"
#include "mem.h"
#include "net.h" /* for dcc functions */

clone_t *main_clone_allow = NULL;

void
add_clone_allow(char *host, int clones)
{
  clone_t *allow;

  allow = leetmalloc(sizeof(clone_t));
  allow->next = main_clone_allow;
  main_clone_allow = allow;

  allow->host = leetstrdup(host);

  allow->clones = clones;
}

int
del_clone_allow(char *host)
{
  clone_t *tmp, **all = &main_clone_allow;

  for (; *all; all = &((*all)->next))
    if (strcasecmp((*all)->host, host) == 0)
    {
      leetfree((*all)->host, strlen((*all)->host) + 1);
      tmp = *all;
      *all = (*all)->next;
      leetfree(tmp, sizeof(clone_t));
      return 1;
    }
  return 0;
}

void
del_all_clone_allows(void)
{
  clone_t *tmp, **all = &main_clone_allow;

  while (*all)
  {
    leetfree((*all)->host, strlen((*all)->host) + 1);
    tmp = *all;
    *all = (*all)->next;
    leetfree(tmp, sizeof(clone_t));
  }
  main_clone_allow = NULL;
}

clone_t *
find_clone_allow(char *host)
{
  clone_t *clone = main_clone_allow;

  for (; clone; clone = clone->next)
    if (strcasecmp(clone->host, host) == 0)
      return clone;
  return NULL;
}

int
tell_clone_allows(sock_t *sock)
{
  clone_t *clone = main_clone_allow;
  user_t *user;
  u_short count;
  int hc;

  if (!clone)
  {
    tosock(sock, "No clone allow records in my database.");
    return 1;
  }
  for (; clone; clone = clone->next)
  {
    /* XXX Highly inefficient! */
    count = 0;
    for (hc = 0; hc < U_HASH_SIZE; hc++)
    {
      for (user = userhash[hc]; user; user = user->next)
      {
        if (IsPerson(user) && strcasecmp(user->host, clone->host) == 0)
          count++;
      }
    }
    tosock(sock, "(%d/%d) %s", count, clone->clones, clone->host); /* change reply to tosock, and to to sock */
  }
  return 1;
}
