/*
*  IRC Services
*
*  Copyright (C) 2003 Services Developers
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of version 1 of the GNU General Public License,
*  as published by the Free Software Foundation.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. The GNU General Public License
*  contains more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* $Id: clone.c,v 1.31 2003/10/31 14:52:17 wcampbel Exp $
*/

#include "irc_string.h"
#include "defines.h"
#include "config.h"
#include "clone.h"
#include "privs.h"
#include "send.h"
#include "mem.h"
#include "net.h"
#include "me.h"
#include "client.h"
#include "main.h"
#include "protected.h"

clone_t *main_clone_t = NULL;

static void cc_send_allow_warnings(char *, clone_t *, clone_t *);
static void cc_construct_nicklist(char *, char *, clone_t *);

/*
 * If the new user's host matches the host of an existing client's host,
 * add the host to our clone hash if it doesn't already exist.
 * If it does exist, increment its clone count and check it...
 * If the new clone count == CLONE_GLINE then call gline() so they die.
 */

int
check_clone(user_t *user, char *host)
{
  char nicklist[BUFSIZE];
  clone_t *clone, *allow;

  if (IsOper(user) || NoKill(user))
    return 0;

  if (check_protected(user))
    return 0;

  if ((clone = find_clone_host(user->host)))
  {
    if ((allow = find_clone_allow(host)))
    {
      if (++clone->clones <= allow->clones)
        return 0;
      if (clone->clones - allow->clones >= me.clone_over)
      {
        tell_chans("\2[CLONES]\2 %d clones from %s (over clone allow limit of %d) - GLINE +%d", clone->clones, clone->host, allow->clones, me.clone_gline_time);
        alldcc("CLONES: %d clones from %s (over clone allow limit of %d) - GLINE +%d", clone->clones, clone->host, allow->clones, me.clone_gline_time);
        gline(NULL, "*", clone->host, me.clone_gline_time, "clones (over clone allow limit)", 0);
        return 1;
      }
      else if (clone->clones - allow->clones > 0)
      {
        tell_chans("\2[CLONES]\2 %d clones from %s (over clone allow limit of %d)", clone->clones, clone->host, allow->clones);
        alldcc("CLONES: %d clones from %s (over clone allow limit of %d)", clone->clones, clone->host, allow->clones);
        clone_allow_warn(user->nick, clone->clones, allow->clones);
        cc_send_allow_warnings(host, clone, allow);
        return 0;
      }
    }
    else /* no allow found */
    {
      if (++clone->clones <= me.clone_limit)
        return 0;
      if (clone->clones < me.clone_gline)
      {
        cc_construct_nicklist(nicklist, host, clone);
        tell_chans("\2[CLONES]\2 %d clones from %s: %s", clone->clones, clone->host, nicklist);
        alldcc("\2[CLONES]\2: %d clones from %s: %s", clone->clones, clone->host, nicklist);
      }
      else if (clone->clones < me.clone_gline + me.clone_limit)
      {
        tell_chans("\2[CLONES]\2 %d clones from %s: GLINE *@%s +%d", clone->clones, clone->host, clone->host, me.clone_gline_time);
        alldcc("CLONES: %d clones from %s: GLINE *@%s +%d", clone->clones, clone->host, clone->host, me.clone_gline_time);
        return gline(NULL, "*", clone->host, me.clone_gline_time, "Clones - over limit", 0);
      }
      else
      {
        tell_chans("\2[CLONES]\2 \2%d\2 clones from %s: GLINE *@%s +%d", clone->clones, clone->host, clone->host, me.fclone_gline_time);
        alldcc("CLONES: %d clones from %s: GLINE *@%s +%d", clone->clones, clone->host, clone->host, me.fclone_gline_time);
        return gline(NULL, "*", clone->host, me.fclone_gline_time, "Flood clones", 0);
      }
    }
  }
  else /* No clone host found */
  {
    add_clone_host(user->host);
  }
  return 0;
}

static void
cc_send_allow_warnings(char *host, clone_t *clone, clone_t *allow)
{
  int hc;
  user_t *user;

  for (hc = 0; hc < U_HASH_SIZE; hc++)
  {
    for (user = userhash[hc]; user; user = user->next)
    {
      if (IsPerson(user) && strcasecmp(user->host, host) == 0)
        clone_allow_warn(user->nick, clone->clones, allow->clones);
    }
  }
}

static void
cc_construct_nicklist(char *nicklist, char *host, clone_t *clone)
{
  user_t *user;
  int hc;

  nicklist[0] = '\0';

  for (hc = 0; hc < U_HASH_SIZE; hc++)
  {
    for (user = userhash[hc]; user; user = user->next)
    {
      if (IsPerson(user) && strcasecmp(user->host, host) == 0)
      {
        clone_warn(user->nick, clone->clones);
        if (nicklist[0] == '\0')
        {
          strlcpy(nicklist, user->nick, BUFSIZE);
        }
        else
        {
          strlcat(nicklist, ", ", BUFSIZE);
          strlcat(nicklist, user->nick, BUFSIZE);
        }
      }
    }
  }
}

void
add_clone_host(char *host)
{
  clone_t *clone;
  /* XXX hash */

  clone = leetmalloc(sizeof(clone_t));
  clone->next = main_clone_t;
  main_clone_t = clone;

  clone->host = leetstrdup(host);
  clone->clones = 1;
}

int
del_clone_host(char *host)
{
  clone_t *tmp, **all = &main_clone_t;

  for (; *all; all = &((*all)->next))
    if (strcmp((*all)->host, host) == 0)
    {
      leetfree((*all)->host, strlen((*all)->host) + 1);
      tmp = *all;
      *all = (*all)->next;
      leetfree(tmp, sizeof(clone_t));
      return 1;
    }
  return 0;
}

int
del_clone_user(user_t *user)
{
  clone_t *clone;

  if (!(clone = find_clone_host(user->host)))
    return 0;

  if (clone->clones-- - 1 < 2)
    return del_clone_host(clone->host);

  return 1;
}

void
del_all_clones(void)
{
  clone_t *tmp, **all = &main_clone_t;

  while (*all)
  {
    leetfree((*all)->host, strlen((*all)->host) + 1);
    tmp = *all;
    *all = (*all)->next;
    leetfree(tmp, sizeof(clone_t));
  }
  main_clone_t = NULL;
}

clone_t *
find_clone_host(char *host)
{
  clone_t *clone = main_clone_t;

  for (; clone; clone = clone->next)
    if (strcasecmp(clone->host, host) == 0)
      return clone;
  return NULL;
}

int
list_clones(char *to)
{
  clone_t *allow, *clone = main_clone_t;
  int x = 0;
  user_t *user;
  int hc;

  for (; clone; clone = clone->next)
  {
    if (clone->clones <= me.clone_limit)
      continue;
    if ((allow = find_clone_allow(clone->host)))
    {
      if (clone->clones <= allow->clones)
        continue;
      reply(to, "(%d) %d clones from %s (over clone allow limit of %d)", ++x,
            clone->clones, clone->host, allow->clones);
      continue;
    }
    reply(to, "(%d) %d clones from %s", ++x, clone->clones, clone->host);
    for (hc = 0; hc < U_HASH_SIZE; hc++)
    {
      for (user = userhash[hc]; user; user = user->next)
      {
        if (IsPerson(user) && strcasecmp(clone->host, user->host) == 0)
        {
          reply(to, "   %s!%s@%s$%s [%s]", user->nick, user->username, user->host, user->realname, user->server->nick);
        }
      }
    }
  }
  if (x < 1)
  {
    return reply(to, "No clones found!");
  }
  return 1;
}

int
kill_clones(void)
{
  clone_t *clone = main_clone_t;
  user_t *user, *tmp;
  int i = 0, x = 0;
  char *host;
  int hc;

  while (clone)
  {
    if (clone->clones <= me.clone_limit || find_clone_allow(clone->host))
    {
      clone = clone->next;
      continue;
    }
    host = leetstrdup(clone->host);
    clone = clone->next;
    for (hc = 0; hc < U_HASH_SIZE; hc++)
    {
      for (user = userhash[hc]; user; )
      {
        if (IsPerson(user) && strcasecmp(user->host, host) == 0)
        {
          if (!check_protected(user))
          {
            toserv(":%s KILL %s :%s!%s (clones - over %d per host limit (%d) - Please see %s)",
                   me.os_nick, user->nick, me.servname, me.os_nick, me.clone_limit, ++x, me.cloneurl);
            tmp = user;
            user = user->next;
            del_user(tmp);
            i++;
          }
        }
        else
          user = user->next;
      }
    }
    tell_chans("\2[CLONES]\2 killed %d clones on %s", x, host);
    alldcc("\2[CLONES]\2: killed %d clones from %s", x, host);
    leetfree(host, strlen(host) + 1);
    x = 0;
  }
  return i;
}
