/*
*  IRC Services
*
*  Copyright (C) 2003 Services Developers
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of version 1 of the GNU General Public License,
*  as published by the Free Software Foundation.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. The GNU General Public License
*  contains more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* $Id: conf.c,v 1.29 2003/11/05 15:19:50 wcampbel Exp $
*/

#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>

#include "irc_string.h"
#include "irc_types.h"
#include "defines.h"
#include "config.h"
#include "privs.h"
#include "flags.h"
#include "conf.h"
#include "mem.h"
#define OPEN_DATABASES
#include "db.h"
#undef OPEN_DATABASES
#include "me.h"
#include "servers.h"
#include "oline.h"
#include "client.h"
#include "snuggle.h"
#include "clone.h"
#include "csnuggle.h" /* we need this for csnuggle */

int
read_db(u_short dbnum)
{
  FILE *db;
  char line[BUFSIZE];
  char *field[10];
  int i = -1, j;

  if (dbnum == 0)
  {
    while (databases[++dbnum])
      if (read_db(dbnum) == -1)
        return dbnum;
    return 0;
  }
  if (!(db = fopen(databases[dbnum], "r")))
    return -1;

  while (fgets(line, BUFSIZE - 1, db))
  {
    if (line[0] == '#')
      continue;
    if (isdigit((int) line[0]))
      j = atoi(strtok(line, ":"));
    else
    {
      j = line[0];
      strtok(line, ":");
    }
    i = -1;
    while ((field[++i] = strtok(NULL, ":")))
      ;
    field[--i][strlen(field[i]) - 1] = '\0';

    switch (j)
    {
      case 'A':
        if (field[0] && field[1] && atoi(field[1]) > 0)
          add_clone_allow(field[0], atoi(field[1]));
        break;
      case 'S':
        if (field[0] && field[1] && field[2] && field[3] && field[4] && field[5] && field[6] &&
            atoi(field[6]) > 0 && field[7] && atoi(field[7]) > 0)
        {
          if (field[8] != NULL)
            add_snuggle(field[0], field[1], field[2], field[3],
                        field[4], field[5], atoi(field[6]),
                        strtol(field[7], NULL, 10),
                        strtol(field[8], NULL, 10));
          else
            add_snuggle(field[0], field[1], field[2], field[3],
                        field[4], field[5], atoi(field[6]),
                        strtol(field[7], NULL, 10), 0);
        }
        break;

      case 'C':
        /* XXX inefficient */
        if (field[0] && field[1] && field[2] && field[3] && atoi(field[3]) > 0
            && field[4] && atol(field[4]) > 0)
          add_csnuggle(field[0], field[1], field[2], atoi(field[3]),
                       strtol(field[4], NULL, 10));
        break;

    }
  }
  fclose(db);
  return 0;
}

int
write_db(u_short dbnum)
{
  FILE *db;
  clone_t *allow;
  snuggle_t *snuggle;
  csnuggle_t *csnuggle;
  extern clone_t *main_clone_allow;
  extern snuggle_t *main_snuggle;
  extern csnuggle_t *main_csnuggle;

  if (dbnum == 0)
  {
    while (databases[++dbnum])
      if (write_db(dbnum) == -1)
        return dbnum;
    return 0;
  }
  if (rename(databases[dbnum], "temp.db") < 0)
    return -1;

  if (!(db = fopen(databases[dbnum], "w")))
    return -1;

  switch (dbnum)
  {
    case DB_ACLONE:
      fprintf(db, "# ACLONE db for %s as of %lu\n", me.servname,
              (unsigned long)time(NULL));
      for (allow = main_clone_allow; allow; allow = allow->next)
        fprintf(db, "A:%s:%d\n", allow->host, allow->clones);
      break;
    case DB_SNUGGLE:
      fprintf(db, "# SNUGGLE db for %s as of %lu\n", me.servname,
              (unsigned long)time(NULL));
      for (snuggle = main_snuggle; snuggle; snuggle = snuggle->next)
        fprintf(db, "S:%s:%s:%s:%s:%s:%s:%d:%lu:%lu\n", snuggle->creator,
                snuggle->nm, snuggle->um, snuggle->hm, snuggle->rm,
                snuggle->reason, snuggle->type,
                (unsigned long)snuggle->created,
                (unsigned long)snuggle->expire);
      break;

    case DB_CSNUGGLE:
      fprintf(db, "# CSNUGGLE db for %s as of %lu\n", me.servname,
              (unsigned long)time(NULL));
      for (csnuggle = main_csnuggle; csnuggle; csnuggle = csnuggle->next)
        fprintf(db, "C:%s:%s:%s:%d:%lu\n", csnuggle->creator, csnuggle->chan,
                csnuggle->reason, csnuggle->type,
                (unsigned long)csnuggle->created);
      break;
  }
  fflush(db);
  fclose(db);
  unlink("temp.db");
  return 0;
}

char
get_flag(u_short flags)
{
  struct privs *pptr;

  for (pptr = privs; pptr->flag != 0; pptr++)
    if (flags & pptr->privs)
      return pptr->flag;
  return '?';
}
