/*
*  IRC Services
*
*  Copyright (C) 2003 Services Developers
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of version 1 of the GNU General Public License,
*  as published by the Free Software Foundation.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. The GNU General Public License
*  contains more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* $Id: csnuggle.c,v 1.34 2003/10/31 15:53:20 wcampbel Exp $
*/

/*
 * Format of csnuggle.db == C:author:#chan:reason:type:time created
 * 5 parvs parv[1]:parv[2]:parv[3]:parv[4]:parv[5]<-TS
 * Do NOT hardcode any #'s for csnuggle->chan ANYPLACE, we dont want,
 * it as hardcoded #'s cause problems(remember snuggle7!!)
 * rm'd - TOP, BOTTOM
 *
 * csnuggle 1 = simply a warning
 * csnuggle 2 = AKILL the client, allowing them to reconnect
 * csnuggle 3 = AGLINE client with an expiry of 4 hours?
 */

#include "irc_types.h"
#include "irc_string.h"
#include "defines.h"
#include "replies.h"
#include "csnuggle.h"
#include "oline.h"
#include "privs.h"
#include "send.h"
#include "mem.h"
#include "md5.h"
#include "me.h"
#include "dcc.h"
#include "client.h"
#include "net.h"
#include "operserv.h"
#include "channels.h"
#include "protected.h"

csnuggle_t *main_csnuggle = NULL;
csnuggle_type csnuggle_types[] = {

                                   { "COUNT", CSNUGGLE_CNT	},
                                   { "ALL", CSNUGGLE_ALL	},
                                   { "MASK", CSNUGGLE_MASK	},
                                   { "CREATOR", CSNUGGLE_CREMASK},
                                   { "ID", CSNUGGLE_ID	},
                                   { (char *)NULL, (u_short)NULL	}
                                 };

/* Original line is char *creator, char *nm, char *um, char *hm, char *rm, char *reason, u_short type, time_t created */

void
add_csnuggle(char *creator, char *chan, char *reason, u_short type, time_t created) /* kllrloop:#opers:drones in channel:1:TS */
{
  csnuggle_t *csnuggle = leetmalloc(sizeof(csnuggle_t));
  csnuggle->next = main_csnuggle;
  main_csnuggle = csnuggle;

  csnuggle->creator = leetstrdup(creator);
  csnuggle->chan = leetstrdup(chan); /* this must use channel mask */
  csnuggle->reason = leetstrdup(reason);

  csnuggle->type = type;
  if (created == 0)
    csnuggle->created = time(NULL);
  else
    csnuggle->created = created;
  csnuggle->matches = 0;
}

int
del_csnuggle(sock_t *sock, char *args)
{
  csnuggle_t *tmp, **all;

  if (args == NULL)
    return NEED_HELP;

  for (all = &main_csnuggle; *all; all = &((*all)->next))
    if (strcasecmp((*all)->chan, args) == 0)
    {
      tell_chans(replies[RPL_CREMCSNUGGLE], sock->name, (*all)->type, (*all)->chan, (*all)->reason);
      leetfree((*all)->creator, strlen((*all)->creator) + 1);
      leetfree((*all)->chan, strlen((*all)->chan) + 1);
      leetfree((*all)->reason, strlen((*all)->reason) + 1);
      tmp = *all;
      *all = (*all)->next;
      leetfree(tmp, sizeof(csnuggle_t));
      return 1;
    }
  return 0;
}

csnuggle_t *
find_csnuggle(char *chan)
{
  csnuggle_t *csnuggle = main_csnuggle;

  for (; csnuggle; csnuggle = csnuggle->next)
    if (strcasecmp(csnuggle->chan, chan) == 0)
      return csnuggle;
  return NULL;
}

int
match_csnuggle(user_t *user, char *chan)
{
  csnuggle_t *csnuggle = main_csnuggle;

  if (IsOper(user) || NoKill(user))
    return 0;

  if (check_protected(user))
    return 0;

  for (; csnuggle; csnuggle = csnuggle->next) /* 1=warn, 2=akill, 3=KLINE */
    if (match(csnuggle->chan, chan))
    {
      csnuggle->matches++;
      switch (csnuggle->type)
      {

          /* Beginnings of channel_snuggle */

        case 1:   /* warn alert channel only */
          if (csnuggle->type == 1 && match(csnuggle->chan, chan))
            csnuggle->matches++;
          tell_chans("\2[CSNUGGLE]\2 %s!%s@%s$%s - %s - %s", user->nick, user->username, user->host, user->realname,
                     csnuggle->reason, csnuggle->chan);
          break;
        case 2:   /* KILL all users enterign channel, exempt irc operators from this */
          tell_chans("\2[CHANKILL]\2 %s!%s@%s$%s - %s - %s",
                     user->nick, user->username,
                     user->host, user->realname,
                     csnuggle->reason,
                     csnuggle->chan);
          irc_kill(user->nick, csnuggle->reason);
          del_user(user);
          return 1;
          break; /* NOT REACHED */

        case 3:   /* GLINE all users entering - expemt services/opers from this as well */
          md5_csnuggle(csnuggle);
          tell_chans("\2[CGLINE]\2 %s!%s@%s$%s - %s - %s",
                     user->nick, user->username,
                     user->host, user->realname,
                     csnuggle->reason,
                     csnuggle->chan);
          gline(NULL, "*", user->host, 12,
                csnuggle->reason, 2);
          /* removed ~*, we will kline *@ */
          return 1;
          break; /* NOT REACHED */

          /* I'm thinking as a case 4, possibly make this the actual =forbid= type of function
           * because i think a channel_jupe could be very effective in its own manner.
           * Note - I think too, i will use the forbid code as this part of csnuggle
           * which will now eliminate .forbid out of DCC, will be a part of .csnuggle
           */

      }
    }
  return 0;
}

char *
md5_csnuggle(csnuggle_t *csnuggle)
{
  char buf[BUFSIZE]; /* need to find what we want to hash thru md5 */
  snprintf(buf, BUFSIZE, "%s:%s:%d", csnuggle->chan, csnuggle->reason, csnuggle->type);
  return wiz_md5(buf);
}

void
del_all_csnuggles(void)
{
  csnuggle_t *tmp, **all = &main_csnuggle;

  while (*all)
  { /* change this to reflect csnuggle items */
    leetfree((*all)->creator, strlen((*all)->creator) + 1);
    leetfree((*all)->chan, strlen((*all)->chan) + 1);
    leetfree((*all)->reason, strlen((*all)->reason) + 1);
    tmp = *all;
    *all = (*all)->next;
    leetfree(tmp, sizeof(csnuggle_t));
  }
  main_csnuggle = NULL;
}

int
list_csnuggles(sock_t *sock, char *args, u_short qtype)
{
  csnuggle_t *csnuggle = main_csnuggle;
  u_short i = 0;

  if (!csnuggle)
    tosock(sock, replies[ERR_NOCSNUGGLES]);

  switch (qtype)
  {
    case CSNUGGLE_CNT:
      for (; csnuggle; csnuggle = csnuggle->next)
        i++;
      tosock(sock, replies[RPL_CSNUGGLECNT], i);
      break;
    case CSNUGGLE_ALL:
      for (; csnuggle; csnuggle = csnuggle->next)
      {
        tosock(sock, replies[RPL_CSNUGLISTA], ++i,
               csnuggle->chan, csnuggle->reason,
               csnuggle->matches);
        tosock(sock, replies[RPL_CSNUGLISTB],
               leetctime(csnuggle->created),
               csnuggle->creator, csnuggle->type);
      }
      tosock(sock, replies[RPL_ENDOFCSNUGL]);
      break;
    case CSNUGGLE_MASK:
      if (!args || args[0] != '#')
        return NEED_HELP;
      for (; csnuggle; csnuggle = csnuggle->next)
        if (match(args, csnuggle->chan))
        {
          tosock(sock, replies[RPL_CSNUGLISTA],
                 ++i, csnuggle->chan,
                 csnuggle->reason,
                 csnuggle->matches);
          tosock(sock, replies[RPL_CSNUGLISTB],
                 leetctime(csnuggle->created),
                 csnuggle->creator,
                 csnuggle->type);
        }
      if (i > 0)
        tosock(sock, replies[RPL_ENDOFCSNUGL]);
      else
        tosock(sock, replies[ERR_NOMATCHCSNUG]);
      break;
    case CSNUGGLE_CRECNT:
      if (!args)
        return NEED_HELP;
      for (; csnuggle; csnuggle = csnuggle->next)
        if (match(args, csnuggle->creator))
          i++;
      tosock(sock, replies[RPL_CRECNTCSNUG], i);
      break;
    case CSNUGGLE_CREMASK:
      if (!args)
        return NEED_HELP;
      for (; csnuggle; csnuggle = csnuggle->next)
        if (match(args, csnuggle->creator))
        {
          tosock(sock, replies[RPL_CSNUGLISTA],
                 ++i, csnuggle->chan,
                 csnuggle->reason,
                 csnuggle->matches);
          tosock(sock, replies[RPL_CSNUGLISTB],
                 leetctime(csnuggle->created),
                 csnuggle->creator,
                 csnuggle->type);
        }
      if (i > 0)
        tosock(sock, replies[RPL_ENDOFCSNUGL]);
      else
        tosock(sock, replies[ERR_NOMATCHCSNUG]);
      break;
    case CSNUGGLE_ID:
      if (!args)
        return NEED_HELP;
      if (strlen(args) != 32 || strstr(args, " "))
      {
        tosock(sock, replies[ERR_CSNUGID]);
        return 1;
      }

      for (csnuggle = main_csnuggle; csnuggle; csnuggle = csnuggle->next)
      {
        if (strcmp(md5_csnuggle(csnuggle), args) == 0)
          break;
      }
      if (csnuggle == NULL)
      {
        /* csnuggle will be NULL when you reach the
        ** end of the list, that means there were
        ** no matches.
        */
        tosock(sock, replies[ERR_CSNUGID]);
        break;
      }

      tosock(sock, replies[RPL_CSNUGLISTA], 1,
             csnuggle->chan, csnuggle->reason,
             csnuggle->matches);
      tosock(sock, replies[RPL_CSNUGLISTB],
             leetctime(csnuggle->created), csnuggle->creator,
             csnuggle->type);
      break;
  }
  return 1;
}

u_short
os_csnuggleid(user_t *cptr, char *to, char *args)
{
  csnuggle_t *csnuggle;

  if (!args)
    return NEED_HELP;
  if (strlen(args) != 32 || strstr(args, " "))
    return reply(cptr->nick, replies[ERR_CSNUGID]);

  for (csnuggle = main_csnuggle; csnuggle; csnuggle = csnuggle->next)
  {
    if (strcmp(md5_csnuggle(csnuggle), args) == 0)
      break;
  }
  if (csnuggle == NULL)
  {
    return reply(cptr->nick, replies[ERR_CSNUGID]);
  }

  reply(cptr->nick, replies[RPL_CSNUGLISTA], 1, csnuggle->chan,
        csnuggle->reason, csnuggle->matches);
  reply(cptr->nick, replies[RPL_CSNUGLISTB], leetctime(csnuggle->created),
        csnuggle->creator, csnuggle->type);
  return 1;
}

