/*
*  IRC Services
*
*  Copyright (C) 2003 Services Developers
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of version 1 of the GNU General Public License,
*  as published by the Free Software Foundation.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. The GNU General Public License
*  contains more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
*  $Id: dcc.c,v 1.125 2003/11/05 22:13:34 kllrloop Exp $
*/

#include <errno.h>
#include "config.h" /* for forbid */
#include "irc_types.h"
#include "irc_string.h"
#include "defines.h"
#include "replies.h"
#include "client.h"
#include "conf.h"
#include "oline.h"
#include "privs.h"
#include "main.h" /* for slog() */
#include "send.h"
#include "dcc.h"
#include "md5.h"
#include "mem.h"
#include "net.h"
#include "rsa.h"
#define OPEN_DATABASES
#include "db.h"
#undef OPEN_DATABASES
#include "me.h"
#include "snuggle.h"
#include "operserv.h"
#include "servers.h"
#include "channels.h"
#include "config.h"
#include "clone.h"
#include "parse.h"
#include "bconf.h"
#include "dlink.h"
#include "csnuggle.h"
#include "serno.h"

#ifdef XLINE
extern int xline(sock_t *, char *, int, char *);
#endif

u_short	dcc_status	(sock_t *, char *);
void	dcc_status_help	(sock_t *, u_short);
u_short	dcc_who	(sock_t *, char *);
void	dcc_who_help	(sock_t *, u_short);
u_short	dcc_whoc	(sock_t *, char *);
void	dcc_whoc_help	(sock_t *, u_short);
u_short	dcc_whokill	(sock_t *, char *);
void	dcc_whokill_help(sock_t *, u_short);
u_short	dcc_save	(sock_t *, char *);
void	dcc_save_help	(sock_t *, u_short);
u_short	dcc_jupe	(sock_t *, char *);
void	dcc_jupe_help	(sock_t *, u_short);
u_short	dcc_rehash	(sock_t *, char *);
void	dcc_rehash_help	(sock_t *, u_short);
u_short	dcc_reload	(sock_t *, char *);
void	dcc_reload_help	(sock_t *, u_short);
u_short	dcc_restart	(sock_t *, char *);
void	dcc_restart_help(sock_t *, u_short);
u_short	dcc_die	(sock_t *, char *);
void	dcc_die_help	(sock_t *, u_short);
u_short	dcc_raw	(sock_t *, char *);
void	dcc_raw_help	(sock_t *, u_short);
u_short	dcc_help	(sock_t *, char *);
u_short	dcc_quit	(sock_t *, char *);
u_short dcc_snuggle	(sock_t *, char *);
void dcc_snuggle_help	(sock_t *, u_short);
u_short dcc_aclone	(sock_t *, char *);
void dcc_aclone_help	(sock_t *, u_short);
u_short dcc_list	(sock_t *, char *);
void dcc_list_help	(sock_t *, u_short);
u_short dcc_cwho	(sock_t *, char *);
void dcc_cwho_help	(sock_t *, u_short);
u_short dcc_kill	(sock_t *, char *);
void dcc_kill_help	(sock_t *, u_short);
u_short dcc_echo	(sock_t *, char *);
void dcc_echo_help	(sock_t *, u_short);
#ifdef XLINE
u_short dcc_xline	(sock_t *, char *);
void dcc_xline_help	(sock_t *, u_short);
#endif
u_short dcc_csnuggle	(sock_t *, char *);
void dcc_csnuggle_help	(sock_t *, u_short);
u_short dcc_settime	(sock_t *, char *);
void dcc_settime_help	(sock_t *, u_short);
u_short dcc_me	(sock_t *, char *);
void dcc_me_help	(sock_t *, u_short);
u_short dcc_killclones	(sock_t *, char *);
void dcc_killclones_help	(sock_t *, u_short);

static struct _dcctab
{
  char *cmd;
  u_short privs;
  u_short (*func)(sock_t *, char *);
  void (*help)(sock_t *, u_short);
}
dcctab[] = {
             { "LOGOUT", 0, dcc_quit, NULL	},
             { "EXIT", 0, dcc_quit, NULL	},
             { "QUIT", 0, dcc_quit, NULL	},
             { "HELP", 0, dcc_help, NULL	},
             { "STATUS", 0, dcc_status, dcc_status_help	},
             { "WHOM", 0, dcc_status, dcc_status_help	},
             { "WHO", 0, dcc_who, dcc_who_help	},
             { "WHOC", 0, dcc_whoc, dcc_whoc_help	},
             { "WHOKILL", 0, dcc_whokill, dcc_whokill_help},
             { "KILL", 0, dcc_kill, dcc_kill_help	},
	     { "KILLCLONES", 0, dcc_killclones, dcc_killclones_help },
             { "SNUGGLE", 0, dcc_snuggle, dcc_snuggle_help},
             { "ACLONE", 0, dcc_aclone, dcc_aclone_help },
             { "ECHO", 0, dcc_echo, dcc_echo_help },
             { "SETTIME", 0, dcc_settime, dcc_settime_help},
             { "JUPE", CADMIN, dcc_jupe, dcc_jupe_help	},
             { "SAVE", CADMIN, dcc_save, dcc_save_help	},
             { "REHASH", CADMIN, dcc_rehash, dcc_rehash_help	},
             { "RELOAD", CADMIN, dcc_reload, dcc_reload_help	},
             { "RESTART", CADMIN, dcc_restart, dcc_restart_help},
             { "DIE", ADMIN, dcc_die, dcc_die_help	},
             { "RAW", ADMIN, dcc_raw, dcc_raw_help	},
#ifdef XLINE
             { "XLINE", ADMIN, dcc_xline, dcc_xline_help },
#endif
             { "CSNUGGLE", CADMIN, dcc_csnuggle, dcc_csnuggle_help},
             { "LIST", CADMIN, dcc_list, dcc_list_help	},
             { "CWHO", CADMIN, dcc_cwho, dcc_cwho_help	},
             { "ME", CADMIN, dcc_me, dcc_me_help },
             { NULL, 0, NULL, NULL	}
           };

void
greet_dcc(sock_t *sock)
{
  tosock(sock, " ");
  tosock(sock, " Welcome to the DCC interface for %s!", me.servname);
  tosock(sock, "You are connecting from %s:%d and your connection has been logged. If you find any bugs, please contribute and or ask why, at http://www.operserv2.org under the FAQ section.",
         inet_ntoa(sock->addr),
         sock->port);
  tosock(sock, " ");
  tosock(sock, " Username");
}

void
parse_dcc(sock_t *sock, char *buffer)
{
  struct _dcctab *dptr = dcctab;
  oline_t *oline;
  u_short wp = 0;
  char *chal;
  char coreLine[BUFSIZE + 1];

  if ((buffer[0] == '\0') || (buffer[0] == '\r') ||
      (buffer[0] == '\n'))
    return ;

  strlcpy(coreLine, buffer, BUFSIZE + 1);

  if (!sock->name)
  {
    sock->name = leetstrdup(strtok(buffer, " "));
    wp = 1;
  }
  if (!(oline = sock->oline) && (!(sock->oline = oline = find_oline_by_nick(sock->name)) || dcc_privs(sock, oline->privs) == 0))
  {
    tosock(sock, "You do not have an o:line for services2, goodbye.");
    sock->flags |= SOCK_DEL;
    return ;
  }
  if (!(sock->flags & SOCK_AUTH))
  {
    if (wp)
    {
#ifdef HAVE_LIBCRYPTO
      if (oline->rsa_public_key && (chal = make_challenge(&sock->response, oline->rsa_public_key)))
      {
        tosock(sock, " ");
        tosock(sock, " Challenge:");
        tosock(sock, "%s", chal);
        tosock(sock, "Response:");
        leetfree(chal, strlen(chal) + 1);
        return ;
      }
      else
      {
#endif
        tosock(sock, " ");
        tosock(sock, " Password");
        return ;
#ifdef HAVE_LIBCRYPTO

      }
#endif

    }
    else
    {
      if ((sock->response && strcasecmp(sock->response, buffer) == 0) ||
          (oline->passwd && strcmp(oline->passwd, wiz_md5(buffer)) == 0))
      {
#if 0
        del_sock(sock); /* remove their "ghost" login, if it exists */
#endif

        sock->flags |= SOCK_AUTH;
        tosock(sock, " ");
        alldcc("*** %s has logged in.", sock->name);
        tell_chans("\2[Login]\2 %s has successfully logged in", sock->name);
        slog("LOGIN by %s on %s", sock->name, inet_ntoa(sock->addr)); /* lets log logins too! */
        if (me.default_echo)
          sock->flags |= SOCK_DCC_ECHO;
        return ;
      }
      else
      {
        tosock(sock, "Denied - Authentication failed.");
        sock->flags |= SOCK_DEL;
        alldcc("Failed login from %s as %s", inet_ntoa(sock->addr), sock->name);
        tell_chans("\2[FAILED]\2 login attempt for %s!", sock->name);
        return ;
      }
    }
  }
  if (*buffer == '.')
  {
    if (strlen(buffer) < 2)
      return ;
    strtok(++buffer, " ");
    for (; dptr->cmd; dptr++)
      if (strcasecmp(dptr->cmd, buffer) == 0)
        break;
    if (dptr->cmd == NULL)
    {
      tosock(sock, "Invalid command.");
      return ;
    }
    if (! (dptr->privs == 0 || (sock->flags & dptr->privs) == dptr->privs))
    {
      tosock(sock, "%s", replies[ERR_NOACCESS]);
      return ;
    }
    /* XXX insert flood checking here */
    tosocks(CADMIN, "#%s# %s %s", sock->name, buffer, (chal = strtok(NULL, "")) ? chal : "");
    if ((*dptr->func)(sock, chal) == NEED_HELP && *dptr->help)
    {
      tosock(sock, "Usage: ");
      (*dptr->help)(sock, HELP_USAGE);
    }
  }
  else
  {
    if (strncasecmp(buffer, "\001ACTION ", 8) == 0)
    {
      char *tptr;

      buffer += 8; /* Get past "\001ACTION " */
      tptr = strchr(buffer, '\001');
      if (tptr != NULL)
        *tptr = '\0';
      dcc_action(sock, buffer);
    }
    else
    {
      dcc_partyline(sock, buffer);
    }
  }
}

u_short
dcc_privs(sock_t *sock, u_short new_privs)
{
  if (new_privs & ADMIN)
    new_privs |= CADMIN;
  if (new_privs & CADMIN)
  {
    new_privs |= SVSOP;
    me.admins++;
  }
  sock->flags |= new_privs;
  return sock->flags;
}

sock_t *
dcc_chat_offer(user_t *cptr)
{
  sock_t *sock = listen_sock();

  toserv(":%s PRIVMSG %s :\1DCC CHAT chat %lu %d\1", me.os_nick, cptr->nick, iptolong((u_int)me.ip.s_addr), sock->port);
  return sock;
}


/* dcc commands */


u_short
dcc_help(sock_t *sock, char *args)
{
  struct _dcctab *dptr = dcctab;

  if (args)
  {
    strtok(args, " ");
    for (; dptr->cmd; dptr++)
      if (strcasecmp(dptr->cmd, args) == 0 && *dptr->help)
      {
        tosock(sock, "Usage: ");
        (*dptr->help)(sock, HELP_USAGE | HELP_DESC);
        return 1;
      }
    tosock(sock, "No help on that command. Try .help for a complete list of commands.");
  }
  else
  {
    for (; dptr->cmd; dptr++)
      if (*dptr->help)
      {
        write(sock->socket, dptr->cmd, strlen(dptr->cmd));
        write(sock->socket, "               ", 15 - strlen(dptr->cmd));
        (*dptr->help)(sock, HELP_DESC);
      }
  }
  return 1;
}

u_short
dcc_save(sock_t *sock, char *args)
{
  short x;

  slog("%s SAVE", sock->name);
  alldcc("%s is manually saving all databases...", sock->name);
  if ((x = write_db(0)) > 0)
  {
    slog("%s SAVE ERROR! %s: %s", sock->name, databases[x], strerror(errno));
    alldcc("ERROR! Unable to write %s: %s", databases[x], strerror(errno));
    return 0;
  }
  alldcc("Database save complete!", sock->name);
  return 1;
}

u_short
dcc_list(sock_t *sock, char *args)
{
  if (!args)
    return NEED_HELP;
  list_channels(sock, args);
  return 1;
}

u_short
dcc_cwho(sock_t *sock, char *args)
{
  channel *chan;

  if (!args)
    return NEED_HELP;
  if (check_chan(sock, args))
    return 1;
  if (!(chan = find_channel(args)))
  {
    tosock(sock, replies[ERR_NOSUCHCHAN], args);
    return 1;
  }

  return channel_who(sock, chan);
}

u_short
dcc_restart(sock_t *sock, char *args)
{
  short x;

  if (!args)
    return NEED_HELP;
  slog("%s RESTART (%s)", sock->name, args);
  alldcc("%s is restarting services! (%s)", sock->name, args);
  if ((x = write_db(0)) > 0)
  {
    slog("%s SAVE ERROR! %s: %s", sock->name, databases[x], strerror(errno));
    alldcc("ERROR! save failed on %s: %s", databases[x], strerror(errno));
    return 0;
  }
  alldcc("restarting.......");
  operwall(OW_CLIENT, "\2RESTART\2 has been requested by \2%s\2", sock->name);
  toserv(":%s QUIT :Restarting!", me.os_nick);
  tell_chans("\2[RESTART]\2 %s told me to restart!", sock->name);
  toserv("ERROR :restart requested");
  fclose(me.logfd);
  close_all_connections();
  execl(me.argv[0], me.argv[0], NULL);
  /* execl() may fail, and return... oh well */
  exit(1);
  /* NOT REACHED */
  return 1;
}

u_short
dcc_jupe(sock_t *sock, char *args)
{
  char buf[BUFSIZE], *server, *reason;

  if (!args || !(server = strtok(args, " ")) || !(reason = strtok(NULL, "")))
    return NEED_HELP;

  if (!strstr(server, "."))
  {
    tosock(sock, "Invalid servername!");
    return 0;
  }
  slog("%s:%s added JUPE: %s (%s)", sock->name, inet_ntoa(sock->addr), server, reason);

  /* Announce the JUPE to DCC,chans and operwall */
  alldcc("%s has JUPED %s! (%s)", sock->name, server, reason);
  operwall(OW_CLIENT, "\2JUPE\2 for \2%s\2 added by Services Admin %s", server, sock->name, reason);
  tell_chans("\2[JUPE]\2 for \2%s\2 added by Services Admin %s", server, sock->name, reason);
  snprintf(buf, BUFSIZE, "JUPED: %s", reason);
  jupe(server, buf);

  return 1;
}

#ifdef XLINE
u_short
dcc_xline(sock_t *sock, char *args)
/* Format is :OperServ2 some.irc.server type(0-2) *gecos* */
{
  char *typec, *gecos, *reason;
  int type;

  if (!args)
    return NEED_HELP;

  typec = strtok(args, " ");
  gecos = strtok(NULL, " ");
  reason = strtok(NULL, "");

  if ((typec == NULL) || (gecos == NULL))
    return NEED_HELP;

  if (reason == NULL)
    reason = "No Reason";

  type = atoi(typec); /* XXX Verification currently being done in xline() */
  if (strcmp(gecos, "*") == 0)
  {
    tosock(sock, "XLINE cannot be be set to \"*\"");
    return NEED_HELP;
  }
  /* XXX Other common abusive XLINE's here */

  slog("%s:%s XLINE(%d) attempt on %s (%s)", sock->name, inet_ntoa(sock->addr),
       type, gecos, reason);
  operwall(OW_CLIENT, "%s \2XLINE(%d)\2 attempt on %s (%s)", sock->name, type,
           gecos, reason);
  tell_chans("\2[XLINE]\2 %s attempted an XLINE(%d) for %s (%s)", sock->name,
             type, gecos, reason);
  xline(sock, gecos, type, reason);
  return 1;
}
#endif

u_short
dcc_die(sock_t *sock, char *args)
{
  short x;

  if (!args)
    return NEED_HELP;
  slog("%s:%s DIE", sock->name, inet_ntoa(sock->addr)); /* log nick, and host */
  alldcc("%s is shutting down services! (%s)", sock->name, args);
  alldcc("Saving databases...");
  if ((x = write_db(0)) > 0)
  {
    slog("%s SAVE ERROR! %s: %s", sock->name, databases[x], strerror(errno));
    alldcc("ERROR! Unable to write %s: %s", databases[x], strerror(errno));
    return 0;
  }
  tell_chans("Dieing...");
  alldcc("Dieing...");
  operwall(OW_CLIENT, "\2DIE\2 requested by \2%s\2", sock->name);
  toserv(":%s QUIT :%s told me to die!", me.os_nick, sock->name);
  close_all_connections();
  fclose(me.logfd);
  exit(1);
  /* NOT REACHED */
  return 1;
}

u_short
dcc_rehash(sock_t *sock, char *args)
{
  alldcc("%s is rehashing...", sock->name);
  slog("%s:%s REHASH", sock->name, inet_ntoa(sock->addr)); /* log nick and host */
  rehash();
  alldcc("\002[REHASH]\002 requested by %s.", sock->name);
  ;
  tell_chans("Rehashing..");
  return 1;
}

u_short
dcc_settime(sock_t *sock, char *args)
{
  if (me.servtype != SERV_IRCNN)
  {
    tosock(sock, "SETTIME will not function on ircd-hybrid.");
    return 1;
  }
  if (me.use_settime == 0)
  {
    tosock(sock, "SETTIME is disabled.");
    return 1;
  }
  slog("%s:%s SETTIME", sock->name, inet_ntoa(sock->addr));
  operwall(OW_CLIENT, "%s requested: SETTIME", sock->name);
  tosock(sock, "clocks set to: %lu", settime(NULL));
  return 1;
}

u_short
dcc_killclones(sock_t *sock, char *args)
{
  u_short i = kill_clones();

  if (i < 1)
  {
    tosock(sock, "No clones found.");
    return 1;
  }
  slog("%s DCC_KILLCLONES - killed %d clones", sock->name, i);
  tell_chans("\2[CLONES]\2 A total of %d clones were killed by %s", i, sock->name);
  tosock(sock, "Killed a total of %d clones.", i);
  return 1;
}

u_short
dcc_reload(sock_t *sock, char *args)
{
  alldcc("%s is manually reloading databases...", sock->name);
  slog("%s:%s told me to RELOAD", sock->name, inet_ntoa(sock->addr));
  if (reload())
  {
    alldcc("ERROR! while reloading databases!");
    return 0;
  }
  alldcc("databases have been reloaded.");
  ;
  return 1;
}

u_short
dcc_raw(sock_t *sock, char *args)
{
  if (!args)
    /* if no args, return help */
    return NEED_HELP;
  if (me.conn == 0)
  {
    tosock(sock, "I'm not connected to a hub right now.");
    return 0;
  }
  tosock(sock, "args = %s", args);
  /* Make sure we LOG any RAW command sent by an oper to uplink */
  slog("%s:%s RAW: %s", sock->name, inet_ntoa(sock->addr), args);
  toserv("%s", args);
  return 1;
}

u_short
dcc_snuggle(sock_t *sock, char *args) /* S:creator:nm:um:hm:rm:cm:reason:type */
{
  char *what, *nm, *um, *hm, *rm, *reason, *tmp;
  snuggle_t *snuggle;
  int type;
  oline_t *oline;
  time_t expire = 0;
  int days = 0;
  char *tocheck;

  if (!args || !(what = strtok(args, " ")))
    return NEED_HELP;

  if (strcasecmp(what, "LIST") == 0)
  {
    snuggle_type *sptr = snuggle_types;

    if ((tmp = strtok(NULL, " ")))
      for (; sptr->name; sptr++)
        if (strcasecmp(sptr->name, tmp) == 0)
        {
          tmp = strtok(NULL, "");
          return list_snuggles(sock, tmp, sptr->qtype); /* change sock->name to sock */
        }
    return NEED_HELP;
  }
  nm = strtok(NULL, "!");
  if (nm == NULL)
    return NEED_HELP;
  if (*nm == '+')
  {
    nm++;
    days = atol(nm);
    tocheck = strchr(nm, ' '); /* Can't use strtok here */
    if (tocheck != NULL)
    {
      nm = tocheck;
      nm++;
    }
    else
      return NEED_HELP;
    if (days > 0)
    {
      /* Though it won't harm too much if it's 0 */
      expire = days * SECINDAY + time(NULL);
    }
  }
  if (!(um = strtok(NULL, "@")) || !(hm = strtok(NULL, "$")) || !(rm = strtok(NULL, ":")))
    return NEED_HELP;

  if (strcasecmp(what, "DEL") == 0)
  {
    if (!(snuggle = find_snuggle(nm, um, hm, rm)))
    {
      tosock(sock, "No matching SNUGGLE records were found.");
      return 0;
    }
    /* add below here of %s and snuggle->chan */
    slog("%s:%s removed SNUGGLE for %s!%s@%s$%s - %s", sock->name, inet_ntoa(sock->addr), nm, um, hm, rm, snuggle->reason);

    /* SNUGGLE DEL code */

    del_snuggle(sock, nm, um, hm, rm);
    if (write_db(DB_SNUGGLE) == -1)
    {
      tell_chans("\002[ERROR!]\002 Unable to write %s: %s", databases[DB_SNUGGLE], strerror(errno));
      tosock(sock, "\002[ERROR!]\002 Unable to write %s: %s", databases[DB_SNUGGLE], strerror(errno));
      return 0;
    }
    tosock(sock, "The matching SNUGGLE was removed successfully.");
    return 1;
  }
  if (!(sock->oline = oline = find_oline_by_nick(sock->name)))
  {
    tosock(sock, "You do not have access to modify the list of SNUGGLE records, please contact an admin!");
    return 0;
  }
  if (!(reason = strtok(NULL, ":")) || !(tmp = strtok(NULL, "")))
    return NEED_HELP;
  if (strcasecmp(what, "ADD") == 0)
  {
    if (check_nm_sock(sock, nm) || check_um_sock(sock, um) || check_hm_sock(sock, hm) || check_rm_sock(sock, rm) || check_reason_sock(sock, reason))
      return 1;

    if ((!(type = atoi(tmp)) && strcmp(tmp, "0") != 0) || type < 1 || type > 6)
    {
      tosock(sock, "Please choose a SNUGGLE type between 1 and 6.");
      return 0;
    }

    if (match(nm, me.os_nick) && match(um, me.os_username) && match(hm, me.servname) &&
        match(rm, me.os_gecos))
    {
      tosock(sock, "You cannot SNUGGLE that.");
      return 0;
    }
    if (type == 3 && check_nick_sock(sock, nm) && (!(sock->oline = oline = find_oline_by_nick(sock->name)))) /* new */
    {
      tosock(sock, "Wildcards are not allowed in a type 3 SNUGGLE nickmask; ask an admin to add it for you.");
      return 1;
    }
    if (type == 6 && um[0] != '~')
    {
      tosock(sock, "For a type 6 SNUGGLE, '~' must be the first character of your username mask.");
      return 1;
    }

    if ((snuggle = find_snuggle(nm, um, hm, rm)))
    {
      char tmp[BUFSIZE];
      snprintf(tmp, BUFSIZE, "%s!%s@%s$%s", snuggle->nm, snuggle->um, snuggle->hm, snuggle->rm);
      tosock(sock, "The following entry already exists and matches your request:");
      return list_snuggles(sock, tmp, SNUGGLE_MASK);
    }
    slog("%s:%s added SNUGGLE %d for %s!%s@%s$%s - %s", sock->name, inet_ntoa(sock->addr), type, nm, um, hm, rm, reason);
    add_snuggle(sock->name, nm, um, hm, rm, reason, type, 0, expire);
    if (expire != 0)
      tell_chans("\2[ASNUGGLE]\2 %s added TEMP(%d day) SNUGGLE(%d) for %s!%s@%s$%s (%s)",
                 sock->name, days, type, nm, um, hm, rm, reason);
    else
      tell_chans("\2[ASNUGGLE]\2 %s added SNUGGLE(%d) for %s!%s@%s$%s (%s)",
                 sock->name, type, nm, um, hm, rm, reason);
    if (write_db(DB_SNUGGLE) == -1)
    {
      tell_chans("ERROR! Unable to write %s: %s", databases[DB_SNUGGLE], strerror(errno));
      tosock(sock, "ERROR! Unable to write %s: %s", databases[DB_SNUGGLE], strerror(errno));
      return 1;
    }
    tosock(sock, "Your SNUGGLE was added successfully.");
    return 0;
  }
  return NEED_HELP;
}

u_short
dcc_csnuggle(sock_t *sock, char *args) /* C:creator:chan:reason:type:TS */
{
  char *chan, *reason, *tmp;
  csnuggle_t *csnuggle;
  int type = 0;
  char *what;

  if (!args || !(what = strtok(args, " "))) /* no args -- return_help */
    return NEED_HELP;

  if (strcasecmp(what, "LIST") == 0)
  {
    csnuggle_type *sptr = csnuggle_types;

    if ((tmp = strtok(NULL, " ")))
      for (; sptr->name; sptr++)
        if (strcasecmp(sptr->name, tmp) == 0)
        {
          return list_csnuggles(sock, tmp, sptr->qtype);
        }
    return NEED_HELP;
  }
  chan = strtok(NULL, "$");
  reason = strtok(NULL, "!");
  tmp = strtok(NULL, " ");
  if (tmp != NULL)
    type = atoi(tmp);

  if (chan == NULL || *chan == '\0')
    return NEED_HELP;

  if (strcasecmp(what, "DEL") == 0)
  {
    if (!(csnuggle = find_csnuggle(chan)))
    {
      tosock(sock, "No matching CSNUGGLE records were found.");
      return 0;
    }
    /* add below here of %s and snuggle->chan */
    slog("%s:%s removed CSNUGGLE for - %s - %s", sock->name, inet_ntoa(sock->addr), csnuggle->chan, csnuggle->reason);
    del_csnuggle(sock, chan);
    if (write_db(DB_CSNUGGLE) == -1)
    {
      tell_chans("\002[ERROR!]\002 Unable to write %s: %s", databases[DB_CSNUGGLE], strerror(errno));
      tosock(sock, "\002[ERROR!]\002 Unable to write %s: %s", databases[DB_CSNUGGLE], strerror(errno));
      return 0;
    }
    tosock(sock, "The matching CSNUGGLE was removed successfully.");
    return 1;
  }

  if (strcasecmp(what, "ADD") == 0)
  {
    if (check_chan(sock, chan)) /* Verify the channel is valid */
      return 1;

    if (reason == NULL)
      return NEED_HELP;

    if ((type < 1) || (type > 3))
    {
      tosock(sock, "Please choose a CSNUGGLE type between 1 and 3.");
      return 0;
    }

    /* XXX Check for my channels */

    if (*chan != '#')
    {
      tosock(sock, "For CSNUGGLES, '#' MUST be in the channel mask.");
      return 1;
    }

    if ((csnuggle = find_csnuggle(chan)))
    {
      char tmp[BUFSIZE];
      snprintf(tmp, BUFSIZE, "%s", csnuggle->chan);
      tosock(sock, "The following entry already exists and matches your request:");
      /* XXX Inefficient */
      return list_csnuggles(sock, tmp, CSNUGGLE_MASK);
    }
    slog("%s:%s added CSNUGGLE %d for %s - %s", sock->name, inet_ntoa(sock->addr), type, chan, reason);
    add_csnuggle(sock->name, chan, reason, type, 0);
    tell_chans("\2[ACSNUGGLE]\2 %s added SNUGGLE(%d) for %s (%s)", sock->name, type, chan, reason);
    if (write_db(DB_CSNUGGLE) == -1)
    {
      tell_chans("ERROR! Unable to write %s: %s", databases[DB_CSNUGGLE], strerror(errno));
      tosock(sock, "ERROR! Unable to write %s: %s", databases[DB_CSNUGGLE], strerror(errno));
      return 1;
    }
    tosock(sock, "Your CSNUGGLE was added successfully.");
    return 0;
  }
  return NEED_HELP;
}

u_short
dcc_aclone(sock_t *sock, char *args)
{
  char *what, *hm, *tmp;
  clone_t *allow;
  int clones = 0;
  oline_t *oline;

  if (!args || !(what = strtok(args, " ")))
    return NEED_HELP;

  if (strcasecmp(what, "LIST") == 0)
    return tell_clone_allows(sock); /* remove sock->name */

  if (!(hm = strtok(NULL, " ")))
    return NEED_HELP;

  if (!(sock->oline = oline = find_oline_by_nick(sock->name)))
  {
    tosock(sock, "You do not have access to modify the list of ACLONE records, please contact an admin.");
    return 1;
  }

  if (strcasecmp(what, "DEL") == 0)
  {
    if (!(allow = find_clone_allow(hm)))
    {
      tosock(sock, "No existing ACLONE entry matched your request.");
      return 1;
    }
    slog("%s:%s called CLONE ALLOW: DEL (%d) for %s", sock->name, inet_ntoa(sock->addr), clones, hm);
    tell_chans("%s removed CLONE ALLOW %s for %d clones.", sock->name, allow->host, allow->clones);
    del_clone_allow(hm);
    if (write_db(DB_ACLONE) == -1)
    {
      tell_chans("ERROR! Unable to write %s: %s", databases[DB_ACLONE], strerror(errno));
      {
        tosock(sock, "ERROR! Unable to write %s: %s", databases[DB_ACLONE], strerror(errno));
        return 1;
      }
    }
    {
      tosock(sock, "The matching ACLONE entry was removed successfully.");
      return 1;
    }
  }
  if ((clones = atoi((tmp = strnull(strtok(NULL, ""))))) < 1 && match("0*", tmp))
    return NEED_HELP;

  if (strcasecmp(what, "ADD") == 0)
  {
    if (check_hm_sock(sock, hm))
      return 1;

    if (clones <= me.clone_limit)
    {
      tosock(sock, "Everyone is already allowed to have at least %d clones!",
             me.clone_limit);
      return 1;
    }
    if (check_host_sock(sock, hm))
      return 1;

    if ((allow = find_clone_allow(hm)))
    {
      tosock(sock, "%s is \2ALREADY\2 allowed %d clones, so no.", allow->host, clones);
      return 1;
    }
    slog("%s:%s CLONE ALLOW: ADD (%d) for %s", sock->name, inet_ntoa(sock->addr), clones, hm);
    tell_chans("\2[ALCONE]\2 %s added CLONE ALLOW %s for %d clones.", sock->name, hm, clones);
    add_clone_allow(hm, clones);
    if (write_db(DB_ACLONE) == -1)
    {
      slog("%s:%s SAVE ERROR! %s", sock->name, inet_ntoa(sock->addr), strerror(errno));
      tell_chans("ERROR! Unable to write %s: %s", databases[DB_ACLONE], strerror(errno));
      {
        tosock(sock, "ERROR! Unable to write %s: %s", databases[DB_ACLONE], strerror(errno));
        return 1;
      }
    }
    tosock(sock, "%s is now allowed to have %d clones.", hm, clones);
    return 0;
  }
  return NEED_HELP;
}

u_short
dcc_kill(sock_t *sock, char *args)
{
  char *who, *reason;
  user_t *user;

  if (!args || !(who = strtok(args, " ")) || !(reason = strtok(NULL, "")))
    return NEED_HELP;

  if (strcasecmp(who, me.os_nick) == 0)
  {
    tosock(sock, "Cannot kill myself.");
    return 1;
  }

  user = find_client(who);
  if (user == NULL)
  {
    tosock(sock, "%s is not a valid user", who);
    return 1;
  }
  if (user->privs != NOPRIV)
  {
    tosock(sock, "I'm not going to kill %s.", who);
    return 1;
  }

  slog("%s:%s requested KILL for %s", sock->name, inet_ntoa(sock->addr), args);
  tell_chans("\2[KILL]\2 requested by %s for %s", sock->name, args); /* do NOT expose IP to channels */
  operwall(OW_CLIENT, "%s requested: KILL for %s (%s)", sock->name, who, reason);
  irc_kill(who, reason);
  del_user(user);
  return 1;
}

u_short
dcc_quit(sock_t *sock, char *args)
{
  if (args)
    alldcc("*** %s has logged out: %s", sock->name, args);
  else
    alldcc("*** %s has logged out.", sock->name);
  sock->flags |= SOCK_DEL;
  return 1;
}

u_short
dcc_who(sock_t *sock, char *args)
{
  char *nm, *um, *hm, *rm;

  if (!args || !(nm = strtok(args, "!")) || !(um = strtok(NULL, "@")) || !(hm = strtok(NULL, "$")) || !(rm = strtok(NULL, "")))
    return NEED_HELP;

  slog("%s:%s WHO: %s!%s@%s$%s", sock->name, inet_ntoa(sock->addr), nm, um, hm, rm);
  if (who(sock, nm, um, hm, rm) == 0)
  {
    tosock(sock, "No clients matched.");
    return 0;
  }
  tosock(sock, "End of WHO.");
  return 1;
}

u_short
dcc_whoc(sock_t *sock, char *args)
{
  char *nm, *um, *hm, *rm;

  if (!args || !(nm = strtok(args, "!")) || !(um = strtok(NULL, "@")) || !(hm = strtok(NULL, "$")) || !(rm = strtok(NULL, "")))
    return NEED_HELP;

  slog("%s:%s WHOC: %s!%s@%s$%s", sock->name, inet_ntoa(sock->addr), nm, um, hm, rm);
  tosock(sock, "%d clients matched %s!%s@%s$%s.",
         whoc(nm, um, hm, rm), nm, um, hm, rm);
  return 1;
}

u_short
dcc_whokill(sock_t *sock, char *args)
{
  char *nm, *um, *hm, *rm, *reason;

  if (!args || !(nm = strtok(args, "!")) || !(um = strtok(NULL, "@")) || !(hm = strtok(NULL, "$")) || !(rm = strtok(NULL, ":")) || !(reason = strtok(NULL, "")))
    return NEED_HELP;

  slog("%s:%s WHOKILL: %s!%s@%s$%s", sock->name, inet_ntoa(sock->addr), nm, um, hm, rm);
  whokill(sock, nm, um, hm, rm, reason);
  return 1;
}

u_short
dcc_echo(sock_t *sock, char *args)
{
  if (args == NULL)
  {
    tosock(sock, "DCC ECHO is \2%s\2",
           (sock->flags & SOCK_DCC_ECHO) ? "ON" : "OFF");
    return 1;
  }
  if (strcasecmp(args, "ON") == 0)
    sock->flags |= SOCK_DCC_ECHO;
  else if (strcasecmp(args, "OFF") == 0)
    sock->flags &= ~SOCK_DCC_ECHO;
  else
    tosock(sock, "Invalid parameter to .echo, please use either "
           "\2ON\2 or \2OFF\2.");

  return 1;
}

u_short
dcc_status(sock_t *sock, char *args)
{
  sock_t *socks;
  dlink_node *dl;
  time_t upt = time(NULL) - me.htmtime;
  snuggle_t *snuggle;
  extern snuggle_t *main_snuggle;
  u_short i = 0;
  csnuggle_t *csnuggle;
  extern csnuggle_t *main_csnuggle;
  u_short j = 0;

  tosock(sock, "%s up %d day%s, %d:%02d:%02d, %d admin%s, memory in use: %d bytes", shtime(), upt / 86400,
         S(upt / 86400), (upt / 3600) % 24, (upt / 60) % 60, upt % 60, me.admins, S(me.admins), me.memusage);
  DLINK_FOREACH(dl, main_sock.head)
  {
    socks = dl->data;

    if (socks->flags & (SOCK_DCC) && !(socks->flags & SOCK_LISTEN))
      tosock(sock, "[%c] %9.9s | idle %d seconds", get_flag(socks->flags), socks->name, time(NULL) - socks->last);
  }
  tosock(sock, "There are %d user%s, %d oper%s, on %d server%s, and %d channel%s formed.", me.users,
         S(me.users), me.opers, S(me.opers), me.servers, S(me.servers), me.channels, S(me.channels));
  tosock(sock, "Serial number: %s", SERIALNUM);
  for (snuggle = main_snuggle; snuggle; snuggle = snuggle->next)
    i++;
  tosock(sock, replies[RPL_SNUGGLECNT], i);
  for (csnuggle = main_csnuggle; csnuggle; csnuggle = csnuggle->next)
    j++;
  tosock(sock, replies[RPL_CSNUGGLECNT], j);

  if (me.conn)
    tosock(sock, "I have been connected to my current HUB %s:%d for %d day%s, %d:%02d:%02d", me.hub->name,
           me.hub->port, upt / 86400, S(upt / 86400), (upt / 3600) % 24, (upt / 60) % 60, upt % 60);
  else
    tosock(sock, "I am not currently connected to any HUB!");
  return 1;
}

void
dcc_status_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".status");
  if (flags & HELP_DESC)
    tosock(sock, "Displays who is logged into %s and other useful information", me.os_nick);
}

void
dcc_who_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".who nm!um@hm$rm");
  if (flags & HELP_DESC)
    tosock(sock, "Displays all clients on the network matching the given mask");
}

void
dcc_whoc_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".whoc nm!um@hm$rm");
  if (flags & HELP_DESC)
    tosock(sock, "Displays the number of clients on the network matching the given mask");
}

void
dcc_whokill_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".whokill nm!um@hm$rm:reason");
  if (flags & HELP_DESC)
    tosock(sock, "Kills all clients on the network matching the given mask, use caution!");
}

void
dcc_settime_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".settime");
  if (flags & HELP_DESC)
    tosock(sock, "Synchronizes time on all servers");
}

void
dcc_kill_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".kill nick reason");
  if (flags & HELP_DESC)
    tosock(sock, "Kills the selected nick.");
}

void
dcc_save_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".save");
  if (flags & HELP_DESC)
    tosock(sock, "Manually saves all services databases");
}

void
dcc_killclones_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
   {
    tosock(sock, ".killclones");
   }
  if (flags & HELP_DESC)
    {
    tosock(sock, "Kills known clones on the network.");
    }
}

void
dcc_snuggle_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
  {
    tosock(sock, "Usage: \2SNUGGLE ADD|LIST|DEL\2 nickmask\2!\2usermask\2@\2hostmask\2$\2namemask:reason:type");
    tosock(sock, "Optional +days argument will set the EXPIREDATE of a SNUGGLE - SNUGGLE ADD +7 nm!um@hm$rm:reason:type");
    tosock(sock, "Also, the +days is not \2required\2.");
    tosock(sock, "There are 6 types of snuggles:");
    tosock(sock, "1: warn the monitoring channel(s)");
    tosock(sock, "2: kill the matching client - Also known as AKILL in other services.");
    tosock(sock, "3: jupe the matching nick - Also known as Nick Jupe.");
    tosock(sock, "4: gline the matching client's username@host - Auto Gline.");
    tosock(sock, "5: gline the matching client's *@host - Also known as AGLINE.");
    tosock(sock, "6: gline the matching client's ~*@host - Very useful for drone/clone attacks.");
    tosock(sock, "Please, use caution on snuggles 5 and 6. \2kthnx!\2");
    tosock(sock, "To retrieve a list of snuggles, use the following:");
    tosock(sock, "\2SNUGGLE LIST all\2 - to retrieve all snuggles");
    tosock(sock, "\2SNUGGLE LIST creator <nick>\2 - to retrieve a list of snuggles created by a certain person");
    tosock(sock, "\2SNUGGLE LIST mask\2 nickmask\2!\2usermask\2@\2hostmask\2$\2namemask");
    tosock(sock, "End of help.");
  }
  if (flags & HELP_DESC)
    tosock(sock, "SNUGGLE matching addresses, nicks, hosts, and gecos. And act according to snuggle-type");
}

void
dcc_aclone_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, "Usage: .aclone add|del host #");
  if (flags & HELP_DESC)
    tosock(sock, "Add or delete clone allow records from my database.");
}
void
dcc_jupe_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".jupe server.name reason");
  if (flags & HELP_DESC)
    tosock(sock, "Forcefully jupes the specified server from the network, remember this server MUST be linked to JUPE.");
}

void
dcc_rehash_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".rehash");
  if (flags & HELP_DESC)
    tosock(sock, "Reloads the services configuration file, and values which can be rehashed");
}

void
dcc_reload_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".reload");
  if (flags & HELP_DESC)
    tosock(sock, "Reloads services databases, reverting to last save");
}

void
dcc_restart_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".restart reason");
  if (flags & HELP_DESC)
    tosock(sock, "Saves all databases and restarts services");
}

void
dcc_die_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".die reason");
  if (flags & HELP_DESC)
    tosock(sock, "Saves all databases and shuts down services");
}

void
dcc_raw_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".raw string");
  if (flags & HELP_DESC)
    tosock(sock, "Sends a raw string to the current hub server- DO NOT ABUSE- kthnx");
}

void
dcc_list_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".list [mask]");
  if (flags & HELP_DESC)
    tosock(sock, "Lists all channels matching the mask provided.");
}

void
dcc_cwho_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".cwho #channel");
  if (flags & HELP_DESC)
    tosock(sock, "Displays all known info for a given channel.");
}

void
dcc_echo_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".echo [ON|OFF]");
  if (flags & HELP_DESC)
    tosock(sock, "Turns local echo on or off, and .echo with no"
           "parameters displays the current echo status.");
}

#ifdef XLINE
void
dcc_xline_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
  {
    tosock(sock, ".xline type *gecos* [reason]");
    tosock(sock, "There are 3 types of XLINES to send:");
    tosock(sock, "0 - Warn opers who have mode +r set.");
    tosock(sock, "1 - Deny the user, and be noisy about it.");
    tosock(sock, "2 - Silently deny the user with matching information");
    tosock(sock, "\2WARNING\2 we must be listed in cluster{} on said server");
  }
  if (flags & HELP_DESC)
  {
    tosock(sock, "Will send a remote XLINE to the server you specify");
  }
}
#endif

void
dcc_csnuggle_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
  {
    tosock(sock, ".csnuggle add|del #chan$reason!type");
    tosock(sock, ".csnuggle list #chanmask");
    tosock(sock, "There are 3 types of CSNUGGLES");
    tosock(sock, "1 - Warn report channel");
    tosock(sock, "2 - AKILL all users entering channel.");
    tosock(sock, "3 - AGLINE all users entering the CSNUGGLED channel");
    tosock(sock, "\2Warning\2 use with caution!!!");
  }
  if (flags & HELP_DESC)
    tosock(sock, "Will SNUGGLE channels on an IRC Network and act as such");
}

u_short
dcc_me(sock_t *sock, char *args)
{
  if (args == NULL)
    return 1;

  dcc_action(sock, args);
  return 1;
}

void
dcc_me_help(sock_t *sock, u_short flags)
{
  if (flags & HELP_USAGE)
    tosock(sock, ".me <message>");
  if (flags & HELP_DESC)
    tosock(sock, "Send an action message to the DCC partyline");
}
