/*
*  IRC Services
*
*  Copyright (C) 2003 Services Developers
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of version 1 of the GNU General Public License,
*  as published by the Free Software Foundation.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. The GNU General Public License
*  contains more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* $Id: net.c,v 1.73 2003/10/23 02:09:37 wcampbel Exp $
*/

#include "irc_types.h"
#include "irc_string.h"
#include "channels.h"
#include "servers.h"
#include "events.h"
#include "config.h"
#include "privs.h"
#include "parse.h"
#include "main.h"
#include "send.h"
#include "dcc.h"
#include "mem.h"
#include "net.h"
#include "me.h"
#include "sclients.h"
#include "dlink.h"

#include <sys/time.h>
#include <assert.h>
#include <errno.h>
#include <stdarg.h>

dlink_list main_sock;
char *servpass = NULL;

#define ERR strerror(errno)

static void login(void);

void
try_next_hub(void)
{
  extern server_t *main_server;
  server_t *next;

  if (me.conn || me.crashing)
    return ;
  if (me.hub != NULL)
    del_sock(me.hub);

  if (!main_server)
  {
    slog("HUB: no C: lines in conf defined! nothing to connect to!");
    return ;
  }
  if (me.lasthub.name)
  {
    if ((next = find_server_t(me.lasthub.name)) && (next = next->next))
    {
      me.lasthub.name = leetrestrdup(me.lasthub.name, next->name);
      me.lasthub.addr.s_addr = inet_addr(next->host);
      me.lasthub.port = next->port;
      servpass = next->pass;
    }
    else
    {
      leetfree(me.lasthub.name, strlen(me.lasthub.name) + 1);
      goto bleh;
    }
  }
  else
  {
bleh:
    me.lasthub.name = leetstrdup(main_server->name);
    me.lasthub.addr.s_addr = inet_addr(main_server->host);
    me.lasthub.port = main_server->port;
    servpass = main_server->pass;
  }
  slog("HUB: autoconnecting to %s on port %d", me.lasthub.name, me.lasthub.port);
  alldcc("HUB: autoconnecting to %s on port %d", me.lasthub.name, me.lasthub.port);
  me.hub = connect_sock(me.lasthub.name, me.lasthub.addr.s_addr, me.lasthub.port, SOCK_HUB);
}

u_short
self_lookup(void)
{
  char hostbuf[BUFSIZE + 1];
  struct hostent *hp;

  gethostname(hostbuf, BUFSIZE);
  hostbuf[BUFSIZE] = 0;
  if (!(hp = gethostbyname((const char *) & hostbuf)))
    return 0;
  memcpy(&me.ip, hp->h_addr_list[0], hp->h_length);
  return 1;
}

sock_t *
add_sock_to_list(void)
{
  sock_t *sock = leetcalloc(sizeof(sock_t), 1);
  dlink_node *dl = dlink_create();

  dlink_add(sock, dl, &main_sock);
  return sock;
}

int
nonblock_sock(int _sock)
{
  int flags, err = 0, sock = (_sock == -1 ? socket(AF_INET, SOCK_STREAM, 0) : _sock);

  if (sock == -1)
    err = slog("NET: unable to allocate socket: %s", ERR);
  else if ((flags = fcntl(sock, F_GETFL, 0)) == -1 || fcntl(sock, F_SETFL, flags | O_NONBLOCK) == -1)
    err = slog("NET: set non-blocking socket failed for %d: %s", sock, ERR);
  return (err ? -1 : sock);
}

sock_t *
connect_sock(char *name, in_addr_t addr, u_short port, u_short flags)
{
  struct sockaddr_in sin;
  int err = 0, _sock;
  sock_t *sock;
  FILE *_fd = NULL;

  memset(&sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = addr;
  sin.sin_port = htons(port);

  if ((_sock = nonblock_sock( -1)) == -1)
    err = 1;
  else if (!(_fd = fdopen(_sock, "r")))
    err = slog("fdopen(%d) failed: %s", _sock, strerror(errno));
  /* add bind() call here */
  else if (connect(_sock, (struct sockaddr *)&sin, sizeof(sin)) == -1 && errno != EINPROGRESS)
  {
    err = slog("%s: connection to %s failed: %s", ((flags & SOCK_HUB) ? "HUB" : "DCC"), name, ERR);
    if (flags & SOCK_HUB)
      alldcc("HUB: connection to %s failed: %s", name, ERR);
  }
  if (err)
  {
    fclose(_fd);
    close(_sock);
    return NULL;
  }
  sock = add_sock_to_list();
  sock->socket = _sock;
  sock->fd = _fd;
  sock->flags = flags;
  sock->addr.s_addr = addr;
  sock->port = port;
  sock->name = leetstrdup(name);
  sock->buffer = leetcalloc(BUFSIZE, 1);
  return sock;
}

sock_t *
listen_sock()
{
  struct sockaddr_in sin;
  int _sock, flags, ospacer = 1;
  sock_t *sock;

  memset(&sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;
  sin.sin_addr.s_addr = htonl(INADDR_ANY);
  sin.sin_port = htons(51000);
  flags = SOCK_DCC;
  if ((_sock = nonblock_sock( -1)) == -1 ||
      setsockopt(_sock, SOL_SOCKET, SO_REUSEADDR, (char *) & ospacer, sizeof(ospacer)) == -1)
    goto err;
  while (bind(_sock, (struct sockaddr *)&sin, sizeof(sin)) == -1)
    if (errno == EADDRINUSE)
      sin.sin_port = htons(ntohs(sin.sin_port) + 1);
    else
      goto err;
  if (listen(_sock, 7) == -1)
    goto err;
  if (0)
  {
err:
    slog("NET: cannot listen on port %d: %s", ntohs(sin.sin_port), ERR);
    close(_sock);
    return NULL;
  }
  sock = add_sock_to_list();
  sock->socket = _sock;
  sock->flags |= (flags | SOCK_LISTEN);
  sock->addr = sin.sin_addr;
  sock->port = ntohs(sin.sin_port);
  sock->last = time(NULL);
  return sock;
}

void
cleanup_dcc(void)
{
  time_t now = time(NULL);
  sock_t *sock;
  dlink_node *dl;
  dlink_node *tdl;

  DLINK_FOREACH_SAFE(dl, tdl, main_sock.head)
  {
    sock = dl->data;

    if (sock->flags & SOCK_DCC)
    {
      if (!(sock->flags & SOCK_CONN) && sock->last + me.conntimeout <= now)
      {
        sock->flags |= SOCK_DEL;
        continue;
      }
      if (!(sock->flags & SOCK_AUTH) && sock->conn + me.authtimeout <= now)
      {
        tosock(sock, "Login grace time exceeded, closing connection.");
        sock->flags |= SOCK_DEL;
      }
    }
  }
}

void
del_sock(sock_t *sock)
{
  dlink_node *dl;

  if (sock->fd)
    fclose(sock->fd);
  shutdown(sock->socket, SHUT_RDWR);
  close(sock->socket);
  dl = dlink_find_delete(sock, &main_sock);
  if (sock->flags & SOCK_HUB && me.crashing != 1)
  {
    slog("HUB: disconnected from %s", sock->name);
    alldcc("HUB: disconnected from %s", sock->name);
    del_all_users();
    add_event(time(NULL) + 7, me.conntimeout / 2, &try_next_hub);
  }
  if (sock->flags & SOCK_AUTH)
  {
    alldcc("*** %s has logged out.", sock->name);
    tell_chans("\2[Logout]\2 %s has logged out", sock->name);
  }
  if (sock == me.hub)
    me.hub = NULL;
  if (sock->flags & CADMIN)
    me.admins--;
  if (sock->name)
    leetfree(sock->name, strlen(sock->name) + 1);
  if (sock->response)
    leetfree(sock->response, strlen(sock->response) + 1);
  leetfree(sock->buffer, BUFSIZE);
  leetfree(sock, sizeof(sock_t));
  dlink_free(dl);
}

void
tosocks(u_short flags, char *tosend, ...)
{
  sock_t *sock;
  dlink_node *dl;
  char msgbuf[BUFSIZE];
  va_list ap;

  if (flags == SOCK_ALL)
    flags = SOCK_AUTH;
  flags |= SOCK_CONN;

  va_start(ap, tosend);
  vsnprintf(msgbuf, BUFSIZE, tosend, ap);
  strlcat(msgbuf, "\n", BUFSIZE);
  DLINK_FOREACH(dl, main_sock.head)
  {
    sock = dl->data;

    if ((sock->flags & flags) == flags)
      write(sock->socket, msgbuf, strlen(msgbuf));
  }
  va_end(ap);
}

void
alldcc(char *tosend, ...)
{
  sock_t *sock;
  dlink_node *dl;
  char msgbuf[BUFSIZE], sendbuf[BUFSIZE], *temp = shtime();
  va_list ap;
  int sendlen;

  va_start(ap, tosend);
  vsnprintf(msgbuf, BUFSIZE, tosend, ap);
  snprintf(sendbuf, BUFSIZE, "(%s) %s\n", temp, msgbuf);
  sendlen = strlen(sendbuf);
  DLINK_FOREACH(dl, main_sock.head)
  {
    sock = dl->data;

    if (sock->flags & SOCK_DEL)
      continue;

    if (sock->flags & (SOCK_DCC) && sock->flags & SOCK_AUTH)
    {
      write(sock->socket, sendbuf, sendlen);
    }
  }
  va_end(ap);
}

void
dcc_partyline(sock_t *sender, char *msg)
{
  sock_t *sock;
  dlink_node *dl;
  char msgbuf[BUFSIZE];

  snprintf(msgbuf, BUFSIZE, "(%s) <%s> %s\n", shtime(), sender->name, msg);
  DLINK_FOREACH(dl, main_sock.head)
  {
    sock = dl->data;

    if (sock->flags & (SOCK_DCC) && sock->flags & SOCK_AUTH &&
        (sock != sender || sender->flags & SOCK_DCC_ECHO))
      write(sock->socket, msgbuf, strlen(msgbuf));
  }
}

void
dcc_action(sock_t *sender, char *msg)
{
  sock_t *sock;
  dlink_node *dl;
  char msgbuf[BUFSIZE];

  if (msg == NULL || *msg == '\0')
    return ;

  snprintf(msgbuf, BUFSIZE, "(%s) * %s %s\n", shtime(), sender->name, msg);
  DLINK_FOREACH(dl, main_sock.head)
  {
    sock = dl->data;

    if (sock->flags & (SOCK_DCC) && sock->flags & SOCK_AUTH &&
        (sock != sender || sender->flags & SOCK_DCC_ECHO))
      write(sock->socket, msgbuf, strlen(msgbuf));
  }
}

void
ping_hub(void)
{
  if (me.conn != 1)
    return ;
  if (me.hub->last + SERVTIMEOUT <= time(NULL))
  {
    slog("HUB: %s is not responding!", me.hub->name);
    alldcc("HUB: %s is not responding!", me.hub->name);
    del_sock(me.hub);
  }
  else
  {
    toserv(":%s PING :%s", me.servname, me.servname);
  }
}

void
io_loop(void)
{
  int r, i, errv;
  socklen_t errlen = sizeof(errv);
  struct timeval tv = { 1, 0 };
  struct sockaddr_in sin;
  size_t ssin = sizeof(sin);
  char buffer[BUFSIZE];
  sock_t *sock, *tmp;
  dlink_node *dl, *tdl;

  memset(buffer, 0, BUFSIZE);
  while (1)
  {
    do_events(time(NULL));

    FD_ZERO(&me.readfds);
    FD_ZERO(&me.writefds);
    DLINK_FOREACH_SAFE(dl, tdl, main_sock.head)
    {
      sock = dl->data;

      if (sock->flags & SOCK_DEL)
      {
        del_sock(sock);
        continue;
      }
      if (sock->flags & (SOCK_LISTEN | SOCK_CONN))
        FD_SET(sock->socket, &me.readfds);
      else
        FD_SET(sock->socket, &me.writefds);
    }

    tv.tv_sec = 1;
    tv.tv_usec = 0;
    r = select(FD_SETSIZE, &me.readfds, &me.writefds, NULL, &tv);
    if (r == -1)
    {
      if (errno != EINTR)
      {
        slog("FATAL! select() == -1: %s", ERR);
        exit(1);
      }
    }
    if (r == 0)
      continue;
    DLINK_FOREACH_SAFE(dl, tdl, main_sock.head)
    {
      sock = dl->data;

      if (FD_ISSET(sock->socket, &me.readfds))
      {
        if (sock->flags & SOCK_LISTEN)
        {
          memset(&sin, 0, sizeof(sin));
          if ((i = accept(sock->socket, (struct sockaddr *) & sin, &ssin)) == -1)
            continue;
          slog("NET: connection from %s:%d", inet_ntoa(sin.sin_addr), ntohs(sin.sin_port));
          tmp = add_sock_to_list();
          if ((tmp->socket = nonblock_sock(i)) == -1)
          {
            del_sock(tmp);
            continue;
          }
          else if (!(tmp->fd = fdopen(tmp->socket, "r")))
          {
            slog("fdopen(%d) failed: %s", tmp->socket, strerror(errno));
            del_sock(tmp);
            continue;
          }
          tmp->conn = time(NULL);
          tmp->addr = sin.sin_addr;
          tmp->port = ntohs(sin.sin_port);
          tmp->buffer = leetcalloc(BUFSIZE, 1);
          tmp->flags |= (SOCK_CONN | sock->flags);
          tmp->flags &= ~SOCK_LISTEN;
          greet_dcc(tmp);
          continue;
        }
        /* Moved up here from after parse()/parse_dcc()
        ** because of the chance of a socket dying and getting
        ** deleted.  -Hwy
        **
        ** Honestly, this isn't TRULY needed at all.
        */
        memset(sock->buffer, 0, BUFSIZE);
        if (!fgets(sock->buffer, BUFSIZE, sock->fd))
        {
          if ((errno == EINTR) || (errno == EAGAIN))
            continue;

          errv = 0;
          if (getsockopt(sock->socket, SOL_SOCKET, SO_ERROR, &errv, &errlen) < 0)
          {
            slog("NET: getsockopt(SO_ERROR) failed: %s", ERR);
            sock->flags |= SOCK_DEL;
            continue;
          }
          sock->flags |= SOCK_DEL;
          slog("NET: read error from %s: %s",
               (sock->name ? sock-> name :
                inet_ntoa(sock->addr)), strerror(errv));
          break;
        }
        sock->last = time(NULL);
        for (i = 0; i < BUFSIZE; i++)
          if (sock->buffer[i] == '\r' || sock->buffer[i] == '\n')
            sock->buffer[i] = 0;
        if (sock->flags & SOCK_HUB)
          parse(sock->buffer);
        else if (sock->flags & (SOCK_DCC))
          parse_dcc(sock, sock->buffer);
      }
      else if (FD_ISSET(sock->socket, &me.writefds) && r--)
      {
        errv = 0;
        if (getsockopt(sock->socket, SOL_SOCKET, SO_ERROR, &errv, &errlen) < 0)
        {
          slog("NET: getsockopt(SO_ERROR) failed: %s", ERR);
          sock->flags |= SOCK_DEL;
          continue;
        }
        if (errv > 0)
        {
          if (sock->flags & SOCK_HUB)
          {
            slog("HUB: error connecting to %s: %s", sock->name, strerror(errv));
            alldcc("HUB: error connecting to %s: %s", sock->name, strerror(errv));
          }
          sock->flags |= SOCK_DEL;
          continue;
        }
        if (sock->flags & SOCK_HUB)
        {
          slog("HUB: connected to %s", sock->name);
          alldcc("HUB: connected to %s", sock->name);
          del_event_func(&try_next_hub);
          sock->conn = time(NULL);
          me.conn = 1;
          login();
        }
        else if (sock->flags & SOCK_DCC)
        {
          slog("DCC: %s connected from %s:%d", sock->name, inet_ntoa(sock->addr), sock->port);
          me.dcc++;
        }
        sock->flags |= SOCK_CONN;
      }
    }
  }
}

void
toserv(char *tosend, ...)
{
  char msgbuf[BUFSIZE + 1];
  va_list ap;

  if (me.conn == 0)
    return ;

  va_start(ap, tosend);
  vsnprintf(msgbuf, BUFSIZE - 2, tosend, ap);
  strlcat(msgbuf, "\r\n", BUFSIZE + 1);
  if (write(me.hub->socket, msgbuf, strlen(msgbuf)) < 0)
  {
    slog("HUB: write error to %s: %s", me.hub->name, ERR);
    alldcc("HUB: write error to %s: %s", me.hub->name, ERR);
    del_sock(me.hub);
    me.conn = 0;
  }
  if (me.debug)
    printf(">>> %s", msgbuf);

  va_end(ap);
}

void
tosock(sock_t *sock, char *tosend, ...)
{
  char msgbuf[BUFSIZE];
  va_list ap;

  if (sock->flags & SOCK_DEL)
    return ;

  va_start(ap, tosend);
  vsnprintf(msgbuf, BUFSIZE, tosend, ap);
  strlcat(msgbuf, "\n", BUFSIZE);
  if (write(sock->socket, msgbuf, strlen(msgbuf)) < 0)
    sock->flags |= SOCK_DEL;
  va_end(ap);
}

void
close_all_connections(void)
{
  sock_t *sock;
  dlink_node *dl, *tdl;

  DLINK_FOREACH_SAFE(dl, tdl, main_sock.head)
  {
    sock = dl->data;
    del_sock(sock); /* XXX inefficient */
  }
}

static void login(void)
{
  time_t now = time(NULL);

  switch (me.servtype)
  {
    case SERV_IRCNN:
      toserv("PASS :%s", servpass);
      toserv("SERVER %s 1 %lu %lu J09 :%s", me.servname,
             now, now, me.servdesc);
      add_linked_server(me.servname, me.servname,
                        me.servdesc);
      intro_nicks();
      join_channels();
      break;
    case SERV_HYBRD:
      toserv("CAPAB :QS EX IE EOB UNKLN KLN HOPS HUB "
             "TBURST CLUSTER");
      toserv("PASS %s :TS", servpass);
      toserv("SERVER %s 0 :%s", me.servname, me.servdesc);
      toserv("SVINFO 5 5 0 :%lu", now);
      add_linked_server(me.servname, me.servname,
                        me.servdesc);
      break;
  }
}
