/*
*  IRC Services
*
*  Copyright (C) 2003 Services Developers
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of version 1 of the GNU General Public License,
*  as published by the Free Software Foundation.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. The GNU General Public License
*  contains more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* $Id: oline.c,v 1.18 2003/09/29 14:27:44 wcampbel Exp $
*/

#include "config.h"
#include "irc_types.h"
#include "irc_string.h"
#include "defines.h"
#include "oline.h"
#include "privs.h"
#include "flags.h"
#include "send.h"
#include "mem.h"
#include "rsa.h"
#include "me.h"
#include "net.h"
#include "dlink.h"
#ifdef	HAVE_LIBCRYPTO
#include <openssl/rsa.h>
#endif

oline_t *main_oline = NULL;

void
add_oline(char *nick, char *passwd, int privs)
{
  oline_t *oline;

  oline = leetcalloc(sizeof(oline_t), 1);
  oline->next = main_oline;
  main_oline = oline;

  oline->nick = leetstrdup(nick);
  oline->privs = privs;
#ifdef	HAVE_LIBCRYPTO

  if (*passwd == '/')
  {
    oline->rsa_public_key_file = leetstrdup(passwd);
    read_public_key(oline);
    oline->passwd = NULL;
    return ;
  }
#endif
  oline->passwd = leetstrdup(passwd);
}

oline_t *
find_oline_by_nick(char *nick)
{
  oline_t *oline = main_oline;

  for (; oline; oline = oline->next)
    if (strcasecmp(oline->nick, nick) == 0)
      return oline;
  return NULL;
}

void
del_all_olines(void)
{
  sock_t *sock;
  dlink_node *dl;
  oline_t *tmp, *all = main_oline;

  DLINK_FOREACH(dl, main_sock.head)
  {
    sock = dl->data;

    sock->oline = NULL;
    sock->flags &= ~(SVSOP | CADMIN | ADMIN);
  }
  while ((tmp = all))
  {
    leetfree(tmp->nick, strlen(tmp->nick) + 1);
#ifdef	HAVE_LIBCRYPTO

    if (tmp->rsa_public_key_file)
      leetfree(tmp->rsa_public_key_file, strlen(tmp->rsa_public_key_file) + 1);
    if (tmp->rsa_public_key)
      RSA_free(tmp->rsa_public_key);
#endif

    all = tmp->next;
    leetfree(tmp, sizeof(oline_t));
  }
  main_oline = NULL;
}

void
stats_o(char *from)
{
  oline_t *oline = main_oline;
  struct privs *pptr;

  for (; oline; oline = oline->next)
  {
    for (pptr = privs; pptr->flag; pptr++)
    {
      if (oline->privs == pptr->privs)
        toserv(":%s 243 %s O *@* * %s %c",
               me.servname, from, oline->nick,
               pptr->flag);
    }
  }
  toserv(":%s 219 %s o :End of /STATS report", me.servname, from);
}
