/*
*  IRC Services
*
*  Copyright (C) 2003 Services Developers
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of version 1 of the GNU General Public License,
*  as published by the Free Software Foundation.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. The GNU General Public License
*  contains more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* $Id: operserv.c,v 1.80 2003/10/31 15:53:20 wcampbel Exp $
*/

#include <unistd.h> /* execl() */
#include <errno.h>
#include "config.h"
#include "irc_types.h"
#include "irc_string.h"
#include "operserv.h"
#include "channels.h"
#include "csnuggle.h"
#include "snuggle.h"
#include "defines.h"
#include "servers.h"
#include "privs.h"
#include "oline.h"
#include "clone.h"
#include "conf.h"
#include "main.h"
#include "send.h"
#include "dcc.h"
#include "mem.h"
#include "md5.h"
#include "net.h"
#include "rsa.h"
#include "me.h"
#include "sclients.h"
#include "dlink.h"
#include "serno.h"
#include "replies.h"
void
chan_clones(user_t *cptr, char *to, char *args)
{
  list_clones(to ? to : cptr->nick);
}

u_short
chan_status(user_t *cptr, char *to, char *args)
{
  snuggle_t *snuggle;
  extern snuggle_t *main_snuggle;
  u_short i = 0;
  time_t upt = time(NULL) - me.htmtime;
  csnuggle_t *csnuggle;
  extern csnuggle_t *main_csnuggle;
  u_short j = 0;

  reply(to ? to : cptr->nick, "%s up %d day%s, %d:%02d:%02d, %d admin%s, memory usage: %d bytes",
        shtime(), upt / 86400, S(upt / 86400), (upt / 3600) % 24, (upt / 60) % 60, upt % 60,
        me.admins, S(me.admins), me.memusage);
  reply(to ? to : cptr->nick, "There are %d user%s, %d oper%s, on %d server%s, and %d channel%s formed.",
        me.users, S(me.users), me.opers, S(me.opers), me.servers, S(me.servers), me.channels, S(me.channels));
  reply(to ? to : cptr->nick, "Serial number: %s", SERIALNUM); /* For debugging reasons */
  for (snuggle = main_snuggle; snuggle; snuggle = snuggle->next)
    i++;
  reply(to ? to : cptr->nick, replies[RPL_SNUGGLECNT], i);
  for (csnuggle = main_csnuggle; csnuggle; csnuggle = csnuggle->next)
    j++;
  reply(to ? to : cptr->nick, replies[RPL_CSNUGGLECNT], j);
  reply(to ? to : cptr->nick, "I have been connected to my current HUB %s:%d for %d day%s, %d:%02d:%02d", me.hub->name,
        me.hub->port, upt / 86400, S(upt / 86400), (upt / 3600) % 24, (upt / 60) % 60, upt % 60);

  return 0;
}

void
chan_killclones(user_t *cptr, char *to, char *args)
{
  u_short i = kill_clones();

  if (i < 1)
  {
    reply(to ? to : cptr->nick, "No clones found.");
    return ;
  }
  slog("%s!%s@%s on %s KILLCLONES killed %d clones", cptr->nick, cptr->username, cptr->host, cptr->server->nick, i);
  tell_chans("\2[CLONES]\2 A total of %d clones were killed by %s", i, cptr->nick);
  reply(cptr->nick, "Killed a total of %d clones.", i);
}

u_short
os_clones(user_t *cptr, char *to, char *args)
{
  chan_clones(cptr, to, args);
  return 1;
}

u_short
os_killclones(user_t *cptr, char *to, char *args)
{
  chan_killclones(cptr, to, args);
  return 1;
}

u_short
os_chat(user_t *cptr, char *to, char *args)
{
  dcc_chat_offer(cptr);
  return 1;
}

u_short
os_mode(user_t *cptr, char *to, char *args)
{
  char *chan, *modes;

  if (!args || !(chan = strtok(args, " ")) || !(modes = strtok(NULL, "")))
    return NEED_HELP;

  if (chan[0] != '#')
    return reply(cptr->nick, "\"%s\" is not a valid channel name.", chan);

  if (!(find_channel(chan)))
    return reply(cptr->nick, "I thought %s was empty?", chan);

  if (modes[0] != '+' && modes[0] != '-')
    return reply(cptr->nick, "\"%s\" is not a valid channel mode", modes);

  slog("%s!%s@%s on %s MODE: %s %s", cptr->nick, cptr->username, cptr->host, cptr->server->nick, chan, modes);
  operwall(OW_CLIENT, "%s!%s@%s on %s requested: MODE %s %s", cptr->nick, cptr->username, cptr->host, cptr->server->nick, chan, modes);
  servmode(chan, modes);
  return 1;
}


u_short
os_global(user_t *cptr, char *to, char *args)
{
  if (!args)
    return NEED_HELP;
  switch (me.servtype)
  {
    case SERV_IRCNN:
      slog("%s!%s@%s on %s requested a global notice be sent :%s", cptr->nick, cptr->username, cptr->host, cptr->server->nick, args);
      operwall(OW_CLIENT, "%s!%s@%s on %s requested a global notice be sent", cptr->nick, cptr->username, cptr->host, cptr->server->nick);

      toserv(":%s NICK GNotice 1 %lu gnotice %s %s :Global Noticer Bot", me.servname, time(NULL), me.servname, me.servname);
      toserv(":GNotice MODE GNOTICE :+iom");
      toserv(":Gnotice JOIN #opers");
      toserv(":GNotice NOTICE $*.* :[Global Message] %s [Do NOT Reply]", args);
      toserv(":GNotice QUIT :Message Sent");
      break;

    case SERV_HYBRD:
      slog("%s!%s@%s on %s requested a global notice be sent :%s", cptr->nick, cptr->username, cptr->host, cptr->server->nick, args);
      operwall(OW_CLIENT, "%s!%s@%s on %s requested a global notice be sent", cptr->nick, cptr->username, cptr->host, cptr->server->nick);

      toserv(":%s NICK GNotice 1 %lu +o gnotice %s %s :Global Noticer Bot", me.servname, time(NULL), me.servname, me.servname);
      toserv(":GNotice MODE GNOTICE :+iom");
      toserv(":Gnotice SJOIN %lu #opers", time(NULL));
      toserv(":GNotice NOTICE $*.* :[Global Message] %s [Do NOT Reply]", args);
      toserv(":GNotice QUIT :Message Sent");
      break;
  }
  return 1;
}


u_short
os_kick(user_t *cptr, char *to, char *args)
{
  user_t *kptr;
  channel *chan;
  char *where, *who, *reason;

  if (!args || !(where = strtok(args, " ")) || !(who = strtok(NULL, " ")) || !(reason = strtok(NULL, "")))
    return NEED_HELP;

  if (is_services_client(who))
    return reply(cptr->nick, "Now why would I want to do that?");

  if (!(chan = find_channel(where)))
    return reply(cptr->nick, "%s is not a valid channel.", where);
  if (!(kptr = find_client(who)) || !del_user_from_channel(kptr, chan))
    return reply(cptr->nick, "%s is not in %s right now.", who, where);

  slog("%s!%s@%s on %s KICK: %s from %s (%s)", cptr->nick, cptr->username, cptr->host, cptr->server->nick, who, where, reason);
  operwall(OW_CLIENT, "%s!%s@%s on %s requested: KICK %s %s (%s)", cptr->nick, cptr->username, cptr->host, cptr->server->nick, where, who, reason);
  toserv(":%s KICK %s %s :%s", me.servname, where, who, reason);
  return 1;
}

u_short
os_settime(user_t *cptr, char *to, char *args)
{
  if (me.servtype != SERV_IRCNN)
  {
    reply(cptr->nick, "SETTIME will not function on ircd-hybrid.");
    return 1;
  }
  if (me.use_settime == 0)
  {
    reply(cptr->nick, "SETTIME is disabled.");
    return 1;
  }
  slog("%s!%s@%s on %s SETTIME", cptr->nick, cptr->username, cptr->host, cptr->server->nick);
  operwall(OW_CLIENT, "%s!%s@%s on %s requested: SETTIME", cptr->nick, cptr->username, cptr->host, cptr->server->nick);
  return reply(cptr->nick, "clocks set to: %lu", settime(NULL));
}

u_short
os_kill(user_t *cptr, char *to, char *args)
{
  char *who, *reason;
  user_t *user;

  if (!args || !(who = strtok(args, " ")) || !(reason = strtok(NULL, "")))
    return NEED_HELP;

  if (strcasecmp(who, me.os_nick) == 0)
    return reply(cptr->nick, "Cannot kill myself.");

  if (!(user = find_client(who)))
    return reply(cptr->nick, "%s: No such nick.", who);

  if (user->privs != NOPRIV)
    return reply(cptr->nick, "I'm not going to kill %s.", who);

  slog("%s!%s@%s on %s KILL: %s (%s)", cptr->nick, cptr->username, cptr->host, cptr->server->nick, who, reason);
  operwall(OW_CLIENT, "%s!%s@%s on %s requested: KILL %s (%s)", cptr->nick, cptr->username, cptr->host, cptr->server->nick, who, reason);
  irc_kill(who, reason);
  del_user(user);
  return 1;
}

u_short
os_placehold(user_t *cptr, char *to, char *args)
{
  char *who, *reason;
  dlink_node *dl;

  if (!args || !(who = strtok(args, " ")) || !(reason = strtok(NULL, "")))
    return NEED_HELP;

  if (check_nick(cptr, who) || check_reason(cptr, reason))
    return 1;

  DLINK_FOREACH(dl, reserved_masks.head)
  {
    if (match(dl->data, who))
      return reply(cptr->nick, "You may not placehold services nicks.");
  }

  if (find_client(who))
    return reply(cptr->nick, "Nickname %s is already in use; cannot be placeheld.", who);

  operwall(OW_CLIENT, "%s!%s@%s on %s requested: PLACEHOLD %s (%s)", cptr->nick, cptr->username, cptr->host, cptr->server->nick, who, reason);
  switch (me.servtype)
  {
    case SERV_IRCNN:
      toserv(":%s NICK %s 1 %lu %s %s %s :%s", me.servname, who, time(NULL), who, me.servname, me.servname, reason);
      toserv(":%s MODE %s :+i", who, who);
      break;
    case SERV_HYBRD:
      toserv(":%s NICK %s 1 %lu +i %s %s %s :%s", me.servname, who, time(NULL), who, me.servname, me.servname, reason);
      break;
  }
  add_user(who, who, me.servname, reason, me.servname, NOPRIV);
  slog("%s!%s@%s on %s PLACEHOLD: %s (%s)", cptr->nick, cptr->username, cptr->host, cptr->server->nick, who, reason);
  return 1;
}

u_short
os_kline(user_t *cptr, char *to, char *args)
{
  return reply(cptr->nick, "KLINE is no longer used, please GLINE instead.");
}

char *
md5_gline(char *um, char *hm, int gtime, char *reason)
{
  char buf[BUFSIZE];
  snprintf(buf, BUFSIZE, "%s@%s %d (%s)", um, hm, gtime, reason);
  return wiz_md5(buf);
}

u_short
os_gline(user_t *cptr, char *to, char *args)
{
  char *um, *hm, *length, *reason;
  int gtime;

  if (!args || !(um = strtok(args, "@")) || !(hm = strtok(NULL, " ")) || !(length = strtok(NULL, " ")) ||
      !(reason = strtok(NULL, "")) || (!(gtime = atoi(length)) && strcmp(length, "0") != 0))
    return NEED_HELP;

  if (gtime < 1 || gtime > 72)
    return reply(cptr->nick, "Please choose a time for your GLINE between 1 and 72 hours.");

  if (check_um(cptr, um) || check_hm(cptr, hm) || check_reason(cptr, reason))
    return 1;

  md5_gline(um, hm, gtime, reason);
  slog("%s!%s@%s on %s GLINE: (ID:%s) %s@%s %d (%s)", cptr->nick,
       cptr->username, cptr->host, cptr->server->nick, me.md5buf,
       um, hm, gtime, reason);
  gline(cptr, um, hm, gtime, reason, 0);
  return 1;
}

u_short
os_ungline(user_t *cptr, char *to, char *args)
{
  char *um, *hm;

  if (me.servtype == SERV_IRCNN)
  {
    reply(cptr->nick, "UNGLINE will not function on ircnn.");
    return 1;
  }

  if (!args)
    return NEED_HELP;

  um = strtok(args, "@");
  hm = strtok(NULL, " ");

  if ((um == NULL) || (hm == NULL))
    return NEED_HELP;

  if (check_um(cptr, um) || check_hm(cptr, hm))
    return 1;

  slog("%s!%s@%s on %s UNGLINE: %s@%s", cptr->nick, cptr->username, cptr->host, cptr->server->nick, um, hm);
  operwall(OW_CLIENT, "%s!%s@%s on %s UNGLINE: %s@%s", cptr->nick, cptr->username, cptr->host, cptr->server->nick, um, hm);
  ungline(cptr, um, hm);
  return 1;
}

u_short
os_anongline(user_t *cptr, char *to, char *args)
{
  char *um, *hm, *length, *reason;
  int gtime;

  if (!args || !(um = strtok(args, "@")) || !(hm = strtok(NULL, " ")) || !(length = strtok(NULL, " ")) ||
      !(reason = strtok(NULL, "")) || (!(gtime = atoi(length)) && strcmp(length, "0") != 0))
    return NEED_HELP;

  if (gtime < 1 || gtime > 72)
    return reply(cptr->nick, "Please choose a time for your ANONGLINE between 1 and 72 hours.");

  if (check_um(cptr, um) || check_hm(cptr, hm) || check_reason(cptr, reason))
    return 1;

  slog("%s!%s@%s on %s ANONGLINE: %s@%s %d (%s)", cptr->nick, cptr->username, cptr->host,
       cptr->server->nick, um, hm, gtime, reason);
  gline(cptr, um, hm, gtime, reason, 1);
  return 1;
}

