/*   IRC Services2
*   Copyright (c) 2003 Services Developers
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of version 1 of the GNU General Public License,
*  as published by the Free Software Foundation.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. The GNU General Public License
*  contains more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* $Id: protected.c,v 1.2 2003/10/31 01:43:36 wcampbel Exp $
*/

#include "irc_string.h"
#include "defines.h"
#include "client.h"
#include "mem.h"
#include "dlink.h"
#include "protected.h"

static dlink_list protected_list;

int check_protected(user_t *u)
{
  dlink_node *dl;
  protected_t *p;

  DLINK_FOREACH(dl, protected_list.head)
  {
    p = dl->data;

    if (match(p->nick, u->nick) && match(p->user, u->username) &&
        match(p->host, u->host) && match(p->gecos, u->realname))
      return 1;
  }

  return 0;
}

void add_protected(char *nick, char *user, char *host, char *gecos)
{
  dlink_node *dl;
  protected_t *p;

  if ((nick == NULL) || (user == NULL) || (host == NULL) || (gecos == NULL))
    return;

  p = leetmalloc(sizeof(protected_t));
  p->nick = leetstrdup(nick);
  p->user = leetstrdup(user);
  p->host = leetstrdup(host);
  p->gecos = leetstrdup(gecos);

  dl = dlink_create();
  dlink_add(p, dl, &protected_list);
}

void del_protected(char *nick, char *user, char *host, char *gecos)
{
  dlink_node *dl, *tdl;
  protected_t *p;

  if ((nick == NULL) || (user == NULL) || (host == NULL) || (gecos == NULL))
    return;  

  /* This doesn't use match, it will ONLY remove a protected entry if it's
  ** exactly what is passed in.
  */
  DLINK_FOREACH_SAFE(dl, tdl, protected_list.head)
  {
    p = dl->data;

    if ((strcmp(p->nick, nick) == 0) && (strcmp(p->user, user) == 0) &&
        (strcmp(p->host, host) == 0) && (strcmp(p->gecos, gecos) == 0))
    {
      dlink_delete(dl, &protected_list);
      leetfree(p->nick, strlen(p->nick) + 1);
      leetfree(p->user, strlen(p->user) + 1);
      leetfree(p->host, strlen(p->host) + 1);
      leetfree(p->gecos, strlen(p->gecos) + 1);
      leetfree(p, sizeof(protected_t));
      dlink_free(dl);

      /* Should I continue here in case something is listed twice?
      ** For now, I'll simply return.
      */
      return;
    } 
  }
}

void flush_protected(void)
{
  dlink_node *dl, *tdl;
  protected_t *p;

  DLINK_FOREACH_SAFE(dl, tdl, protected_list.head)
  {
    p = dl->data;

    dlink_delete(dl, &protected_list);

    leetfree(p->nick, strlen(p->nick) + 1);
    leetfree(p->user, strlen(p->user) + 1);
    leetfree(p->host, strlen(p->host) + 1);
    leetfree(p->gecos, strlen(p->gecos) + 1);
    leetfree(p, sizeof(protected_t));
    dlink_free(dl);
  }
}
