/*
*  IRC Services
*
*  Copyright (C) 2003 Services Developers
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of version 1 of the GNU General Public License,
*  as published by the Free Software Foundation.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. The GNU General Public License
*  contains more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* $Id: send.c,v 1.40 2003/10/23 02:43:39 wcampbel Exp $
*/

#include "irc_types.h"
#include "irc_string.h"
#include "defines.h"
#include "config.h"
#include "privs.h"
#include "send.h"
#include "net.h"
#include "me.h"
#include "main.h"

#include <stdarg.h>

void
is_an_oper_notice(char *server, char *sendbuf, ...)
{
  char notice[BUFSIZE];
  user_t *user;
  va_list ap;
  int hc;

  va_start(ap, sendbuf);
  vsnprintf(notice, BUFSIZE, sendbuf, ap);
  for (hc = 0; hc < U_HASH_SIZE; hc++)
  {
    for (user = userhash[hc]; user; user = user->next)
    {
      if (IsOper(user) && strcmp(user->server->nick, server) != 0)
        toserv(":%s NOTICE %s :*** Notice -- %s", me.servname, user->nick, notice);
    }
  }

  va_end(ap);
  return ;
}

time_t
settime(char *serv)
{
  time_t now = time(NULL);
  user_t *user;
  int hc;

  /* We should not get in here, but just in case... */
  if (me.use_settime == 0)
  {
    slog("WARNING:  settime() called when me.use_settime == 0!");
    return time(NULL);
  }

  if (serv != NULL)
  {
    toserv(":%s SETTIME %lu :%s", me.os_nick, now, serv);
    tell_chans("\2[AUTOTIME]\2 %s clock set to %lu on rejoin", serv, now);
    return now;
  }

  /* else, NULL was passed so you want ALL servers set. */

  for (hc = 0; hc < U_HASH_SIZE; hc++)
  {
    for (user = userhash[hc]; user; user = user->next)
    {
      if (IsServer(user))
        toserv(":%s SETTIME %lu :%s", me.os_nick, now, user->nick);
    }
  }

  return now;
}

int
tell_chans(char *msg, ...)
{
  char msgbuf[BUFSIZE];
  va_list ap;
  int i;

  va_start(ap, msg);
  vsnprintf(msgbuf, BUFSIZE, msg, ap);

  for (i = 0; me.chans[i]; i++)
    toserv(":%s PRIVMSG %s :%s", me.os_nick, me.chans[i], msgbuf);

  va_end(ap);
  return 1;
}

int
reply(char *to, char *notice, ...)
{
  char msgbuf[BUFSIZE];
  va_list ap;

  va_start(ap, notice);
  vsnprintf(msgbuf, BUFSIZE, notice, ap);
  if (to[0] == '=')
    /* This is not supposed to happen, this was for mimicing
    ** ircii DCC.  abort() will simply core it rather than
    ** try to send the message.
    ** XXX
    */
    abort();
  else if (to[0] == '#')
    toserv(":%s PRIVMSG %s :%s", me.os_nick, to, msgbuf);
  else
    toserv(":%s NOTICE %s :%s", me.os_nick, to, msgbuf);

  va_end(ap);
  return 1;
}

void
serv_notice(char *to, char *notice, ...)
{
  char msgbuf[BUFSIZE];
  va_list ap;

  va_start(ap, notice);
  vsnprintf(msgbuf, BUFSIZE, notice, ap);
  toserv(":%s NOTICE %s :%s", me.servname, to, msgbuf);

  va_end(ap);
}

void
clone_warn(char *to, u_int clones)
{
  serv_notice(to, "An excessive number of clones (%d) has been detected from your host. "
              "If you do not limit your connections to 3, your host will be banned from the network. "
              "If you would like to apply for an exemption to the clone rule, please visit %s - "
              "This may be your only warning before you are forcefully removed from this network.",
              clones, me.cloneurl);
}

void
clone_allow_warn(char *to, u_int clones, u_int allowed)
{
  serv_notice(to, "An excessive number of clones (%d) has been detected from your host. "
              "This is over your current allowed number of clones (%d). "
              "Please contact an active operator to request that your limit be raised, "
              "or limit your connections to the amount you are allowed. "
              "If you ignore this warning, your host may be banned from the network.",
              clones, allowed);
}

void
operwall(u_short type, char *operwall, ...)
{
  char wallbuf[BUFSIZE];
  va_list ap;

  va_start(ap, operwall);
  vsnprintf(wallbuf, BUFSIZE, operwall, ap);
  switch (me.servtype)
  {
    case SERV_IRCNN:
      toserv(":%s OPERWALL :%s", (type == OW_SERVER ? me.servname : me.os_nick),
             wallbuf);
      break;
    case SERV_HYBRD:
      if (type == OW_SERVER)
        toserv(":%s WALLOPS :%s", me.servname, wallbuf);
      else
        toserv(":%s OPERWALL :%s", me.os_nick, wallbuf);
      break;
  }
  va_end(ap);
}

void
servmode(char *chan, char *modes)
{
#if 1
  /*
   * with hyb7/ircnn we can just do this
   * with hyb6 we have to get the ts, sjoin, mode, part, blah
   */
  switch (me.servtype)
  {
    case SERV_IRCNN:
      toserv(":%s MODE %s %s", me.servname, chan, modes);
      break;
    case SERV_HYBRD:
      toserv(":%s MODE %s %s", me.os_nick, chan, modes);
      break;
  }
#else
  channel *channel;
  u_short i;

  switch (me.servtype)
  {
    case SERV_IRCNN:
      toserv(":%s MODE %s %s", me.servname, chan, modes);
      break;
    case SERV_HYBRD:
      channel = find_channel(chan);
      for (i = 0; me.chans[i]; i++)
        if (strcasecmp(me.chans[i], chan) == 0)
          goto there;
      if (!channel)
        return ;
      /* :local.host SJOIN 1021703401 #stuff +tn :@irc6 */
      toserv(":%s SJOIN %lu %s %s :@%s", me.servname, channel->ts, channel->name,
             get_channel_modes(channel), me.os_nick);
there:
      assert(0 != channel);
      toserv(":%s MODE %s %s", me.os_nick, channel->name, modes);
      if (!me.chans[i])
        toserv(":%s PART %s", me.os_nick, channel->name);
      break;
  }
#endif
}

void
hackops(char *chan, char *nick)
{
  toserv(":%s MODE %s +o %s", me.servname, chan, nick);
}

void
irc_kill(char *who, char *reason)
{
  toserv(":%s KILL %s :%s!%s (%s)", me.os_nick, who, me.servname, me.os_nick, reason);
}

void
squit(char *server, char *reason)
{
  switch (me.servtype)
  {
    case SERV_IRCNN:
      toserv(":%s SQUIT %s %lu :%s", me.servname, server, time(NULL), reason);
      break;
    case SERV_HYBRD:
      toserv(":%s SQUIT %s :%s", me.servname, server, reason);
      break;
  }
}
