/*   IRC Services2
*   Copyright (c) 2003 Services Developers
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of version 1 of the GNU General Public License,
*  as published by the Free Software Foundation.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. The GNU General Public License
*  contains more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* $Id: services.c,v 1.1 2003/10/31 14:52:18 wcampbel Exp $
*/

#include "irc_string.h"
#include "defines.h"
#include "mem.h"
#include "dlink.h"
#include "services.h"

static dlink_list services_list;

int is_services(char *serv)
{
  dlink_node *dl;
  char *s;

  DLINK_FOREACH(dl, services_list.head)
  {
    s = dl->data;

    if (match(s, serv))
      return 1;
  }

  return 0;
}

void add_services(char *services)
{
  dlink_node *dl;
  char *s;

  if (services == NULL)
    return;

  dl = dlink_create();
  s = leetstrdup(services);
  dlink_add(s, dl, &services_list);
}

void del_services(char *services)
{
  dlink_node *dl;
  char *s;

  if (services == NULL)
    return;  

  DLINK_FOREACH(dl, services_list.head)
  {
    s = dl->data;

    if (strcmp(s, services) == 0)
    {
      dlink_delete(dl, &services_list);
      leetfree(s, strlen(s) + 1);
      dlink_free(dl);
      return;
    } 
  }
}

void flush_services(void)
{
  dlink_node *dl, *tdl;
  char *s;

  DLINK_FOREACH_SAFE(dl, tdl, services_list.head)
  {
    s = dl->data;

    dlink_delete(dl, &services_list);

    leetfree(s, strlen(s) + 1);
    dlink_free(dl);
  }
}
