/*
*  IRC Services
*
*  Copyright (C) 2003 Services Developers
*
*  This program is free software; you can redistribute it and/or modify
*  it under the terms of version 1 of the GNU General Public License,
*  as published by the Free Software Foundation.
*
*  This program is distributed in the hope that it will be useful, but
*  WITHOUT ANY WARRANTY; without even the warranty of MERCHANTABILITY
*  or FITNESS FOR A PARTICULAR PURPOSE. The GNU General Public License
*  contains more details.
*
*  You should have received a copy of the GNU General Public License
*  along with this program; if not, write to the Free Software
*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*
* $Id: snuggle.c,v 1.58 2003/11/04 17:03:38 wcampbel Exp $
*/

#include "config.h" /* Always put config.h first */
#include "irc_types.h"
#include "irc_string.h"
#include "defines.h"
#include "replies.h"
#include "snuggle.h"
#include "oline.h"
#include "privs.h"
#include "send.h"
#include "mem.h"
#include "md5.h"
#include "me.h"
#include "dcc.h"
#include "client.h"
#include "net.h"
#include "operserv.h"
#include "main.h"
#include "protected.h"
#include "db.h"
#include "conf.h"

snuggle_t *main_snuggle = NULL;

snuggle_type snuggle_types[] = {
                                 { "COUNT", SNUGGLE_CNT	},
                                 { "ALL", SNUGGLE_ALL	},
                                 { "MASK", SNUGGLE_MASK	},
                                 { "CRECNT", SNUGGLE_CRECNT	},
                                 { "CREATOR", SNUGGLE_CREMASK	},
                                 { "TOP", SNUGGLE_TOP	},
                                 { "BOTTOM", SNUGGLE_BOTTOM	},
                                 { "ID", SNUGGLE_ID	},
                                 { (char *)NULL, (u_short)NULL	}
                               };

void /* Added in char *chan at an attempt to add snuggle 7 */
add_snuggle(char *creator, char *nm, char *um, char *hm, char *rm, char *reason, u_short type, time_t created, time_t expire)
{
  snuggle_t *snuggle = leetmalloc(sizeof(snuggle_t));
  snuggle->next = main_snuggle;
  main_snuggle = snuggle;

  snuggle->creator = leetstrdup(creator);
  snuggle->nm = leetstrdup(nm);
  snuggle->um = leetstrdup(um);
  snuggle->hm = leetstrdup(hm);
  snuggle->rm = leetstrdup(rm);
  snuggle->reason = leetstrdup(reason);
  snuggle->expire = expire;
  /* if (type == 7) */
  /* snuggle->chan = leetstrdup(chan); added for type 7 snuggles */

  snuggle->type = type;
  if (created == 0)
    snuggle->created = time(NULL);
  else
    snuggle->created = created;
  snuggle->matches = 0;
}

int
del_snuggle(sock_t *sock, char *nm, char *um, char *hm, char *rm)
{
  snuggle_t *tmp, **all;

  for (all = &main_snuggle; *all; all = &((*all)->next))
    if (strcasecmp((*all)->nm, nm) == 0 && strcasecmp((*all)->um, um) == 0 && strcasecmp((*all)->hm, hm) == 0 && strcasecmp((*all)->rm, rm) == 0)
    {
      tell_chans(replies[RPL_CREMSNUGGLE], sock->name, (*all)->type, nm, um, hm, rm, (*all)->reason);
      leetfree((*all)->creator, strlen((*all)->creator) + 1);
      leetfree((*all)->nm, strlen((*all)->nm) + 1);
      leetfree((*all)->um, strlen((*all)->um) + 1);
      leetfree((*all)->hm, strlen((*all)->hm) + 1);
      leetfree((*all)->rm, strlen((*all)->rm) + 1);
      leetfree((*all)->reason, strlen((*all)->reason) + 1);
      /* if ((*all)->type == 7) */
      /* leetfree((*all)->chan, strlen((*all)->chan) +1); free the deleted channel snuggle too! */
      tmp = *all;
      *all = (*all)->next;
      leetfree(tmp, sizeof(snuggle_t));
      return 1;
    }
  return 0;
}

snuggle_t *
find_snuggle(char *nm, char *um, char *hm, char *rm)
{
  snuggle_t *snuggle = main_snuggle;

  for (; snuggle; snuggle = snuggle->next)
    if (strcasecmp(snuggle->nm, nm) == 0 && strcasecmp(snuggle->um, um) == 0 && strcasecmp(snuggle->hm, hm) == 0 && strcasecmp(snuggle->rm, rm) == 0)
      return snuggle;
  return NULL;
}

int
match_snuggle(user_t *user, char *username, char *host, char *realname)
{
  /* char *chan; */
  snuggle_t *snuggle = main_snuggle;

  if (IsOper(user) || NoKill(user))
    return 0;

  if (check_protected(user))
    return 0;

  for (; snuggle; snuggle = snuggle->next)
    if (match(snuggle->nm, user->nick) && (snuggle->type == 3 || (match(snuggle->hm, host) && match(snuggle->um, username) && match(snuggle->rm, realname))))
    {
      snuggle->matches++;
      switch (snuggle->type)
      {
        case 5:
          md5_snuggle(snuggle);
          gline(NULL, "*", host, 12, snuggle->reason, 2);
          tell_chans("\2[AGLINE]\2 %s!%s@%s$%s - %s - %s!%s@%s$%s", user->nick, username, host, realname,
                     snuggle->reason, snuggle->nm, snuggle->um, snuggle->hm, snuggle->rm);
          return 1;
        case 6:
          md5_snuggle(snuggle);
          gline(NULL, "~*", host, 12, snuggle->reason, 2);
          tell_chans("\2[AGLINE]\2 %s!%s@%s$%s - %s - %s!%s@%s$%s", user->nick, username, host, realname,
                     snuggle->reason, snuggle->nm, snuggle->um, snuggle->hm, snuggle->rm);
          return 1;
        case 4:
          md5_snuggle(snuggle);
          gline(NULL, username, host, 12, snuggle->reason, 2);
          tell_chans("\2[AGLINE]\2 %s!%s@%s$%s - %s - %s!%s@%s$%s", user->nick, username, host, realname,
                     snuggle->reason, snuggle->nm, snuggle->um, snuggle->hm, snuggle->rm);
          return 1;
        case 3:
          {
            char nick[BUFSIZE]; /* XXX size */

            strlcpy(nick, user->nick, BUFSIZE);

            irc_kill(user->nick, snuggle->reason);
            del_user(user);
            switch (me.servtype)
            {
              case SERV_IRCNN:
                toserv(":%s NICK %s 1 %lu %s %s %s :%s", me.servname, nick, time(NULL) - 60, nick, me.servname, me.servname, snuggle->reason);
                toserv(":%s MODE %s +i", nick, nick);
                break;
              case SERV_HYBRD:
                toserv(":%s NICK %s 1 %lu +i %s %s %s :%s", me.servname, nick, time(NULL) - 60, nick, me.servname, me.servname, snuggle->reason);
                break;
            }
            add_user(nick, nick, me.servname, snuggle->reason, me.servname, NOPRIV);
            tell_chans("\2[NJUPE]\2 %s!%s@%s$%s - %s - %s!%s@%s$%s", nick, username, host, realname,
                       snuggle->reason, snuggle->nm, snuggle->um, snuggle->hm, snuggle->rm);
            return 1;
          }
        case 2:
          {
            char nick[BUFSIZE];

            strlcpy(nick, user->nick, BUFSIZE);
            irc_kill(nick, snuggle->reason);
            del_user(user);
            tell_chans("\2[AKILL]\2 %s!%s@%s$%s - %s - %s!%s@%s$%s", nick, username, host, realname,
                       snuggle->reason, snuggle->nm, snuggle->um, snuggle->hm, snuggle->rm);
            return 1;
          }
        case 1:
          if (me.eob)
          {
            tell_chans("\2[SNUGGLE]\2 %s!%s@%s$%s - %s - %s!%s@%s$%s", user->nick, username, host, realname,
                       snuggle->reason, snuggle->nm, snuggle->um, snuggle->hm, snuggle->rm);
            /* Don't return here, we want to pass through and continue
            ** checking for any potential non-warning snuggle.  You MAY
            ** end up with multiple notices for a client, but it's not that
            ** much harm.
            **
            ** After alpha, we will move snuggle(1)'s to their own list that
            ** is matched after any non-warning snuggle.
            */
          }
          break;
      }
    }
  return 0;
}

char *
md5_snuggle(snuggle_t *snuggle)
{
  char buf[BUFSIZE];
  snprintf(buf, BUFSIZE, "%s!%s@%s$%s:%d", snuggle->nm, snuggle->um, snuggle->hm, snuggle->rm, snuggle->type);
  return wiz_md5(buf);
}

void
del_all_snuggles(void)
{
  snuggle_t *tmp, **all = &main_snuggle;

  while (*all)
  {
    leetfree((*all)->creator, strlen((*all)->creator) + 1);
    leetfree((*all)->nm, strlen((*all)->nm) + 1);
    leetfree((*all)->um, strlen((*all)->um) + 1);
    leetfree((*all)->hm, strlen((*all)->hm) + 1);
    leetfree((*all)->rm, strlen((*all)->rm) + 1);
    leetfree((*all)->reason, strlen((*all)->reason) + 1);
    /* if ((*all)->type == 7) */
    /* leetfree((*all)->chan, strlen((*all)->chan) + 1); */
    tmp = *all;
    *all = (*all)->next;
    leetfree(tmp, sizeof(snuggle_t));
  }
  main_snuggle = NULL;
}

void
expire_snuggles(void)
{
  snuggle_t *tmp, *all;
  time_t now = time(NULL);
  snuggle_t *prev = NULL;

  all = main_snuggle;
  while (all != NULL)
  {
    if ((all->expire != 0) && (all->expire <= now))
    {
      tell_chans("\2[SNUGGLE EXPIRE]\2 SNUGGLE(%d) for %s!%s@%s$%s (%s)",
                 all->type, all->nm, all->um, all->hm, all->rm, all->reason);
      leetfree(all->creator, strlen(all->creator) + 1);
      leetfree(all->nm, strlen(all->nm) + 1);
      leetfree(all->um, strlen(all->um) + 1);
      leetfree(all->hm, strlen(all->hm) + 1);
      leetfree(all->rm, strlen(all->rm) + 1);
      leetfree(all->reason, strlen(all->reason) + 1);
      if (prev == NULL)
        main_snuggle = all->next;
      else
        prev->next = all->next;
      tmp = all;
      all = all->next;
      leetfree(tmp, sizeof(snuggle_t));
    }
    else
    {
      prev = all;
      all = all->next;
    }
  }
  write_db(DB_SNUGGLE);
}
      
int
list_snuggles(sock_t *sock, char *args, u_short qtype) /* remove char *from > sock_t *sock */
{
  snuggle_t *snuggle = main_snuggle;
  char *nm, *um, *hm, *rm;
  u_short i = 0;

  if (!snuggle)
    tosock(sock, replies[ERR_NOSNUGGLES]);

  switch (qtype)
  {
    case SNUGGLE_CNT:
      for (; snuggle; snuggle = snuggle->next)
        i++;
      tosock(sock, replies[RPL_SNUGGLECNT], i);
      break;
    case SNUGGLE_ALL:
      for (; snuggle; snuggle = snuggle->next)
      {
        tosock(sock, replies[RPL_SNUGLISTA], ++i, snuggle->nm, snuggle->um,
               snuggle->hm, snuggle->rm, snuggle->reason, snuggle->matches);
        tosock(sock, replies[RPL_SNUGLISTB], leetctime(snuggle->created),
               snuggle->creator, snuggle->type);
      }
      tosock(sock, replies[RPL_ENDOFSNUGL]);
      break;
    case SNUGGLE_MASK:
      if (!((nm = strtok(args, "!")) && (um = strtok(NULL, "@")) &&
            (hm = strtok(NULL, "$")) && (rm = strtok(NULL, ""))))
        return NEED_HELP;
      for (; snuggle; snuggle = snuggle->next)
        if (match(nm, snuggle->nm) && match(um, snuggle->um) &&
            match(hm, snuggle->hm) && match(rm, snuggle->rm))
        {
          tosock(sock, replies[RPL_SNUGLISTA], ++i, snuggle->nm,
                 snuggle->um, snuggle->hm, snuggle->rm, snuggle->reason,
                 snuggle->matches);
          tosock(sock, replies[RPL_SNUGLISTB], leetctime(snuggle->created),
                 snuggle->creator, snuggle->type);
        }
      if (i > 0)
        tosock(sock, replies[RPL_ENDOFSNUGL]);
      else
        tosock(sock, replies[ERR_NOMATCHSNUG]);
      break;
    case SNUGGLE_CRECNT:
      if (!args)
        return NEED_HELP;
      for (; snuggle; snuggle = snuggle->next)
        if (match(args, snuggle->creator))
          i++;
      tosock(sock, replies[RPL_CRECNTSNUG], i);
      break;
    case SNUGGLE_CREMASK:
      if (!args)
        return NEED_HELP;
      for (; snuggle; snuggle = snuggle->next)
        if (match(args, snuggle->creator))
        {
          tosock(sock, replies[RPL_SNUGLISTA], ++i, snuggle->nm,
                 snuggle->um, snuggle->hm, snuggle->rm, snuggle->reason,
                 snuggle->matches);
          tosock(sock, replies[RPL_SNUGLISTB], leetctime(snuggle->created),
                 snuggle->creator, snuggle->type);
        }
      if (i > 0)
        tosock(sock, replies[RPL_ENDOFSNUGL]);
      else
        tosock(sock, replies[ERR_NOMATCHSNUG]);
      break;
    case SNUGGLE_TOP:
      if (!args || atoi(args) <= 0)
        return NEED_HELP;
      for (; snuggle; snuggle = snuggle->next)
        i++;
      if (atoi(args) > i)
        tosock(sock, replies[ERR_TOOMANYSNUG], i);
      {
        snuggle_t **all = leetcalloc(i + 1, sizeof(snuggle_t *));
        short j = i = 0;
        for (snuggle = main_snuggle; (all[i++] = snuggle); snuggle = snuggle->next)
          ;
        qsort(all, --i, sizeof(snuggle_t *), &snugtop);
        for (i = atoi(args); j < i && (snuggle = all[j]);)
        {
          tosock(sock, replies[RPL_SNUGLISTA], ++j, snuggle->nm, snuggle->um,
                 snuggle->hm, snuggle->rm, snuggle->reason, snuggle->matches);
          tosock(sock, replies[RPL_SNUGLISTB], leetctime(snuggle->created),
                 snuggle->creator, snuggle->type);
        }
        leetfree(all, (--j * sizeof(snuggle_t *)));
      }
      tosock(sock, replies[RPL_ENDOFSNUGL]);
      break;
    case SNUGGLE_BOTTOM:
      if (!args || atoi(args) <= 0)
        return NEED_HELP;
      for (; snuggle; snuggle = snuggle->next)
        i++;
      if (atoi(args) > i)
        tosock(sock, replies[ERR_TOOMANYSNUG], i);
      {
        snuggle_t **all = leetcalloc(i + 1, sizeof(snuggle_t *));
        short j = i = 0;
        for (snuggle = main_snuggle; (all[i++] = snuggle); snuggle = snuggle->next)
          ;
        qsort(all, --i, sizeof(snuggle_t *), &snugbottom);
        for (i = atoi(args); j < i && (snuggle = all[j]);)
        {
          tosock(sock, replies[RPL_SNUGLISTA], ++j, snuggle->nm, snuggle->um,
                 snuggle->hm, snuggle->rm, snuggle->reason, snuggle->matches);
          tosock(sock, replies[RPL_SNUGLISTB], leetctime(snuggle->created),
                 snuggle->creator, snuggle->type);
        }
        leetfree(all, (--j * sizeof(snuggle_t *)));
      }
      tosock(sock, replies[RPL_ENDOFSNUGL]);
      break;
    case SNUGGLE_ID:
      if (!args)
        return NEED_HELP;
      if (strlen(args) != 32 || strstr(args, " "))
      {
        tosock(sock, replies[ERR_SNUGIDA]);
        return 1;
      }

      for (snuggle = main_snuggle; snuggle; snuggle = snuggle->next)
      {
        if (strcmp(md5_snuggle(snuggle), args) == 0)
          break;
      }
      if (snuggle == NULL)
      {
        /* snuggle will be NULL when you reach the
        ** end of the list, that means there were
        ** no matches.
        */
        tosock(sock, replies[ERR_SNUGIDB]);
        break;
      }

      tosock(sock, replies[RPL_SNUGLISTA], 1, snuggle->nm, snuggle->um, snuggle->hm, snuggle->rm,
             snuggle->reason, snuggle->matches);
      tosock(sock, replies[RPL_SNUGLISTB], leetctime(snuggle->created), snuggle->creator, snuggle->type);
      break;
  }
  return 1;
}

u_short
os_snuggleid(user_t *cptr, char *to, char *args)
{
  snuggle_t *snuggle;

  if (!args)
    return NEED_HELP;
  if (strlen(args) != 32 || strstr(args, " "))
    return reply(cptr->nick, replies[ERR_SNUGIDA]);

  for (snuggle = main_snuggle; snuggle; snuggle = snuggle->next)
  {
    if (strcmp(md5_snuggle(snuggle), args) == 0)
      break;
  }
  if (snuggle == NULL)
  {
    return reply(cptr->nick, replies[ERR_SNUGIDB]);
  }

  reply(cptr->nick, replies[RPL_SNUGLISTA], 1, snuggle->nm,
        snuggle->um, snuggle->hm, snuggle->rm,
        snuggle->reason, snuggle->matches);
  reply(cptr->nick, replies[RPL_SNUGLISTB], leetctime(snuggle->created),
        snuggle->creator, snuggle->type);
  return 1;
}

int
snugtop(const void *a, const void *b)
{
  snuggle_t *c = (snuggle_t *) * (unsigned *)a, *d = (snuggle_t *) * (unsigned *)b;
  return d->matches - c->matches;
}

int
snugbottom(const void *a, const void *b)
{
  snuggle_t *c = (snuggle_t *) * (unsigned *)a, *d = (snuggle_t *) * (unsigned *)b;
  return c->matches - d->matches;
}
